from .optimizer import Optimizer
from .cross_validation import CrossValidationEstimator
from .ensemble_optimizer import EnsembleOptimizer
from .fit_methods import fit, available_fit_methods
from .io import _read_pickle as read_summary

__all__ = ['fit',
           'read_summary',
           'available_fit_methods',
           'Optimizer',
           'EnsembleOptimizer',
           'CrossValidationEstimator']



# TODO: Docstring all functions
""" This module aims to collect debug functions for various parts of the core

"""
from ase.build import bulk
from ase.calculators.emt import EMT
from ase.neighborlist import NeighborList

from matplotlib import pyplot as plt
import itertools
import numpy as np
import spglib as spg


from .atoms import Atom
from .get_clusters import get_clusters
from ..force_constant_model import ForceConstantModel
from ..cluster_space import ClusterSpace
from .utilities import BiMap
from .tensors import rotation_to_cart_coord, rotate_tensor


def debug_fcm_clusters(fcm, atoms, cutoffs):
    fcm_cluster_list = fcm.cluster_list
    debug_cluster_list = get_clusters(atoms, cutoffs, len(atoms))
#    assert len(fcm_cluster_list) == len(debug_cluster_list), \
#            (len(fcm_cluster_list), len(debug_cluster_list))
    for i, c_fcm in enumerate(fcm_cluster_list):
        c_fcm = sorted(c_fcm)
        if c_fcm not in debug_cluster_list:
            assert False, str(c_fcm) + ' not in debug list'
        for j, c_fcm2 in enumerate(fcm_cluster_list):
            c_fcm2 = sorted(c_fcm2)
            if i == j:
                continue
            assert c_fcm != c_fcm2, (i, j)
    print('All clusters seem to be found!')


def simple_test():
    prim = bulk('Al')
    cs = ClusterSpace(prim, [5, 5])
    atoms = bulk('Al', cubic=True).repeat(3)
    fcm = ForceConstantModel(atoms=atoms, cs=cs)
    atoms.calc = EMT()
    pos = atoms.positions.copy()
    atoms.rattle(0.01)
    disps = atoms.positions - pos
    F = atoms.get_forces().flatten()
    M = fcm.get_fit_matrix(disps)
    a = np.linalg.lstsq(M, F)[0]
    plt.plot(F, np.dot(M, a),  '.')
    plt.savefig('test.pdf')
    return (prim, cs, atoms, fcm)


def validate_cs(cs):

    # Use spglib to try to find a primitive cell
    spg_prim = spg.standardize_cell(cs.primitive_structure, no_idealize=True,
                                    to_primitive=True, symprec=cs.symprec)

    # is the cell a valid cell?
    det = np.linalg.det(cs.primitive_structure.cell.T)
    assert det > 1e-12

    # is the cell the same size as spglibs?
    assert np.isclose(det, np.linalg.det(spg_prim[0].T))

    # is the basis compatible?
    assert sorted(cs.primitive_structure.numbers) == sorted(spg_prim[2])

    assert cs.spacegroup == spg.get_spacegroup(cs.primitive_structure,
                                               symprec=cs.symprec)

    # Is the prim spos within prim cell?
    for spos in cs.primitive_structure.basis:
        for s in spos:
            assert s >= 0 and s < 1-1e-4

    # Did the cs find the correct amount of atoms?
    cutoff = cs.cutoffs.max_cutoff
    nl = NeighborList([cutoff / 2] * len(cs.primitive_structure), skin=0,
                      self_interaction=True, bothways=True)
    nl.update(cs.primitive_structure)
    atom_list = BiMap()
    for i in range(len(cs.primitive_structure)):
        for site, offset in zip(*nl.get_neighbors(i)):
            atom = Atom(site, *offset)
            if atom not in atom_list:
                atom_list.append(atom)
    assert len(cs.atom_list) == len(atom_list)
    for atom in atom_list:
        assert atom in cs.atom_list
    for atom in cs.atom_list:
        assert atom in atom_list

    # Is the first atoms in the atom list the center atoms?
    for i in range(len(cs.primitive_structure)):
        atom = cs.atom_list[i]
        assert atom.site == i and tuple(atom.offset) == (0, 0, 0)

    # Does all clusters contain any of the center atoms?
    prim_set = set(range(len(cs.primitive_structure)))
    for cluster in cs.cluster_list:
        assert not prim_set.isdisjoint(cluster)

    # Does the clusters obey the cutoff
    spos = [atom.spos(cs.primitive_structure.basis) for atom in cs.atom_list]
    pos = [np.dot(spos, cs.primitive_structure.cell) for sp in spos]
    for cluster in cs.cluster_list:
        order = len(cluster)
        nbody = len(set(cluster))
        if nbody == 1:
            continue
        cutoff = cs.cutoffs.get_cutoff(order=order, nbody=nbody)
        for i, j in itertools.combinations(set(cluster), r=2):
            assert np.linalg.norm(pos[i] - pos[j]) < cutoff

    for orbit in cs.orbits:
        prototype_index = orbit.prototype_index
        assert (orbit.orientation_families[0].cluster_indices[0] ==
                prototype_index)
        perm = orbit.orientation_families[0].permutation_indices[0]
        assert cs.permutations[perm] == tuple(range(orbit.order))
        prototype_cluster = cs.cluster_list[prototype_index]
        order = len(prototype_cluster)
        assert orbit.order == order
        prototype_distances = []
        for i, j in itertools.combinations(set(prototype_cluster), r=2):
            distance = np.linalg.norm(pos[i] - pos[j])
            assert distance < orbit.maximum_distance
            prototype_distances.append(distance)
        prototype_distances = sorted(prototype_distances)
        for of in orbit.orientation_families:
            cluster_index = of.cluster_indices[0]
            cluster = cs.cluster_list[cluster_index]
            order = len(cluster)
            assert orbit.order == order
            distances = []
            for i, j in itertools.combinations(set(cluster), r=2):
                distance = np.linalg.norm(pos[i] - pos[j])
                assert distance < orbit.maximum_distance
                distances.append(distance)
            distances = sorted(distances)
            assert np.allclose(distances, prototype_distances)

    for orbit in cs.orbits:
        prototype = cs.cluster_list[orbit.prototype_index]
        assert sorted(prototype) == list(prototype)
        values, positions, counts = np.unique(prototype, return_index=True,
                                              return_counts=True)
        for p, c in zip(positions, counts):
            perm = list(range(len(prototype)))
            for tmp in itertools.permutations(perm[p:p+c]):
                perm = list(range(len(prototype)))
                perm[p:p+c] = tmp
                for et in orbit.eigentensors:
                    assert np.allclose(et, et.transpose(perm))

    for orbit in cs.orbits:
        for ri, pi in orbit.eigensymmetries:
            R_scaled = cs.rotation_matrices[ri]
            R = rotation_to_cart_coord(R_scaled, cs.primitive_structure.cell)
            R_inv = np.linalg.inv(R)
            perm = cs.permutations[pi]
            for et in orbit.eigentensors:
                assert np.allclose(rotate_tensor(et, R_inv).transpose(perm),
                                   et)

    nirred = cs._cvs.shape[1]
    irred = np.random.random(nirred)
    params = cs._map_parameters(irred)
    i = 0
    for orb in cs.orbits:
        for of in orb.orientation_families:
            fc = np.zeros([3]*orb.order)
            for j, et in enumerate(of.eigentensors):
                fc += et * params[i+j]
            of.force_constant = fc
        i += len(orb.eigentensors)

    lookup = dict()
    for orbit in cs.orbits:
        for of in orbit.orientation_families:
            fc = of.force_constant
            for ci, pi in zip(of.cluster_indices, of.permutation_indices):
                cluster = cs.cluster_list[ci]
                perm = cs.permutations[pi]
                lookup[tuple(cluster)] = fc.transpose(perm)

#    previous_cluster_prefix = None
    for order in cs.cutoffs.orders:
        for pre in range(len(cs.primitive_structure)):
            for body in itertools.combinations_with_replacement(
                    range(pre, len(cs.atom_list)), r=order-2):
                cluster = [pre] + list(body) + [None]
                FC = np.zeros([3]*order)
                for i in range(len(cs.atom_list)):
                    cluster[-1] = i
                    inv_perm = np.argsort(np.argsort(cluster))
                    fc = lookup.get(tuple(sorted(cluster)))
                    if fc is None:
                        continue
                    FC += fc.transpose(inv_perm)
                assert np.allclose(FC, 0), FC.round(8)

    for orbit in cs.orbits:
        prototype = cs.cluster_list[orbit.prototype_index]
        FC = lookup[tuple(prototype)]
        for of in orbit.orientation_families:
            R_scaled = cs.rotation_matrices[of.symmetry_index]
            R = rotation_to_cart_coord(R_scaled, cs.primitive_structure.cell)
            R_inv = np.linalg.inv(R)
            assert np.allclose(rotate_tensor(FC, R_inv), of.force_constant)



"""
This module introduces the ForceConstantPotential object which acts as the
finalized force constant model.
"""

import pickle
import numpy as np

from collections import Counter

from .force_constant_model import ForceConstantModel
from .core.orbits import Orbit
from .core.orbits import OrientationFamily
from .core.tensors import rotation_to_cart_coord, rotate_tensor


# TODO: Fix the relation with cluster space
class ForceConstantPotential:
    """ A finalized force constant model. Can produce force constants for any
    structure compatible with the structure for which the model was set up.

    Parameters
    ----------
    cs : ClusterSpace
        The cluster space the model is based upon
    parameters : numpy.ndarray
        The fitted paramteres
    metadata : dict
        metadata dictionary, will be pickled when object is written to file
    """

    def __init__(self, cs, parameters, metadata=None):

        self._prim = cs.primitive_structure.copy()
        self.cluster_list = cs.cluster_list.copy()
        self.atom_list = cs.atom_list.copy()
        self.orbits = []
        self.spacegroup = cs.spacegroup
        self._config = cs._config

        # add metadata
        if metadata is None:
            metadata = dict()
        self._metadata = metadata
        self._add_default_metadata()

        # Extract the eigentensors from the cluster space and use the paramters
        # to construct the finalized force constants
        parameters = cs._map_parameters(parameters)
        p = 0
        for orb in cs.orbits:
            new_orbit = Orbit()
            fc = np.zeros(orb.eigentensors[0].shape)
            for et, a in zip(orb.eigentensors, parameters[p:]):
                fc += et * a
            new_orbit.force_constant = fc
            new_orbit.order = orb.order
            new_orbit.radius = orb.radius
            new_orbit.maximum_distance = orb.maximum_distance
            for of in orb.orientation_families:
                new_of = OrientationFamily()
                new_of.cluster_indices = of.cluster_indices.copy()
                sym_ind = of.symmetry_index
                R = rotation_to_cart_coord(cs.rotation_matrices[sym_ind],
                                           self.primitive_structure.cell)
                fc = rotate_tensor(new_orbit.force_constant, R.T)
                perm = cs.permutations[of.permutation_indices[0]]
                new_of.force_constant = fc.transpose(perm)
                new_orbit.orientation_families.append(new_of)
            self.orbits.append(new_orbit)
            p += len(orb.eigentensors)

    @property
    def symprec(self):
        return self._config['symprec']

    @staticmethod
    def read(f):
        """Reads a force constant potential from file.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to load from (file object)

        Returns
        -------
        ForceConstantPotential
            the original object as stored in the file
        """
        if isinstance(f, str):
            with open(f, 'rb') as fobj:

                # TODO: Remove this once better read/write functions in place
                # This allows for reading FCPs with ASE-3.17 and 3.18
                from .core.atoms import Atoms
                fcp = pickle.load(fobj)
                _prim = fcp._prim

                if hasattr(_prim, '_cell'):  # 3.17
                    cell = _prim._cell
                    pbc = _prim._pbc
                else:                       # 3.18
                    cell = _prim.cell[:]
                    pbc = _prim.pbc

                new_prim = Atoms(
                    symbols=_prim.symbols, positions=_prim.positions, cell=cell, pbc=pbc)
                fcp._prim = new_prim
                return fcp
        else:
            try:
                return pickle.load(f)
            except Exception:
                raise Exception('Failed loading from file.')

    def write(self, f):
        """Writes a force constant potential to file.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to write to (file object)
        """
        if isinstance(f, str):
            with open(f, 'wb') as fobj:
                pickle.dump(self, fobj)
        else:
            try:
                pickle.dump(self, f)
            except Exception:
                raise Exception('Failed writing to file.')

    @property
    def metadata(self):
        """ dict : metadata associated with force constant potential """
        return self._metadata

    @property
    def primitive_structure(self):
        """ ase.Atoms : atomic structure """
        return self._prim.copy()

    @property
    def orbit_data(self):
        """ list : list of dictionaries containing detailed information for
                   each orbit, e.g. cluster radius and force constant
        """
        data = []
        for orbit_index, orbit in enumerate(self.orbits):
            d = {}
            d['index'] = orbit_index
            d['order'] = orbit.order
            d['radius'] = orbit.radius
            d['maximum_distance'] = orbit.maximum_distance
            d['n_clusters'] = len(orbit.orientation_families)

            types = []
            for atom_ind in self.cluster_list[orbit.prototype_index]:
                types.append(self.primitive_structure.numbers[
                    self.atom_list[atom_ind].site])
            d['prototype_cluster'] = self.cluster_list[orbit.prototype_index]
            d['prototype_atom_types'] = types

            d['geometrical_order'] = len(set(d['prototype_cluster']))
            d['force_constant'] = orbit.force_constant
            d['force_constant_norm'] = np.linalg.norm(orbit.force_constant)
            data.append(d)
        return data

    def get_force_constants(self, atoms):
        """ Return the force constants of a compatible structure.

        Parameters
        ----------
        atoms : ase.Atoms
            input structure

        Returns
        -------
        ForceConstants
            force constants
        """
        return ForceConstantModel(atoms, self).get_force_constants()

    def __str__(self):
        orbits = self.orbit_data
        orbit_counts = Counter([orbit['order'] for orbit in orbits])
        cluster_counts = Counter()
        for orbit in orbits:
            cluster_counts[orbit['order']] += orbit['n_clusters']

        n = 54
        s = []
        s.append(' ForceConstantPotential '.center(n, '='))
        s.append('Spacegroup {}'.format(self.spacegroup))
        s.append('Cell:\n{}'.format(self.primitive_structure.cell))
        s.append('Basis:\n{}'.format(self.primitive_structure.basis))
        s.append('Numbers: {}'.format(self.primitive_structure.numbers))
        for order in sorted(orbit_counts.keys()):
            s.append('Order {}, #orbits {}, #cluster {}'.format(
                     order, orbit_counts[order], cluster_counts[order]))
        s.append('Total number of orbits: {} '.format(len(orbits)))
        s.append('total number of clusters: {} '
                 .format(sum(cluster_counts.values())))
        s.append(''.center(n, '='))
        return '\n'.join(s)

    def __repr__(self):
        return 'ForceConstantPotential(ClusterSpace({!r}, ...), [...])'.format(
            self.primitive_structure)

    def _add_default_metadata(self):
        """Adds default metadata to metadata dict."""
        import getpass
        import socket
        from datetime import datetime
        from . import __version__ as hiphive_version

        self._metadata['date_created'] = datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
        self._metadata['username'] = getpass.getuser()
        self._metadata['hostname'] = socket.gethostname()
        self._metadata['hiphive_version'] = hiphive_version



"""
Helper functions for reading and writing objects to tar files
"""

import pickle
import tempfile
import warnings
import ase.io as aseIO
with warnings.catch_warnings():
    warnings.filterwarnings('ignore', category=FutureWarning)
    import h5py


def add_ase_atoms_to_tarfile(tar_file, atoms, arcname, format='json'):
    """ Adds an ase.Atoms object to tar_file """
    temp_file = tempfile.NamedTemporaryFile()
    aseIO.write(temp_file.name, atoms, format=format)
    temp_file.seek(0)
    tar_info = tar_file.gettarinfo(arcname=arcname, fileobj=temp_file)
    tar_file.addfile(tar_info, temp_file)


def read_ase_atoms_from_tarfile(tar_file, arcname, format='json'):
    """ Reads ase.Atoms from tar file """
    temp_file = tempfile.NamedTemporaryFile()
    temp_file.write(tar_file.extractfile(arcname).read())
    temp_file.seek(0)
    atoms = aseIO.read(temp_file.name, format=format)
    return atoms


def add_items_to_tarfile_hdf5(tar_file, items, arcname):
    """ Add items to one hdf5 file """
    temp_file = tempfile.NamedTemporaryFile()
    hf = h5py.File(temp_file.name)
    for key, value in items.items():
        hf.create_dataset(key, data=value, compression='gzip')
    hf.close()
    temp_file.seek(0)
    tar_info = tar_file.gettarinfo(arcname=arcname, fileobj=temp_file)
    tar_file.addfile(tar_info, temp_file)
    temp_file.close()


def add_items_to_tarfile_pickle(tar_file, items, arcname):
    """ Add items by pickling them """
    temp_file = tempfile.TemporaryFile()
    pickle.dump(items, temp_file)
    temp_file.seek(0)
    tar_info = tar_file.gettarinfo(arcname=arcname, fileobj=temp_file)
    tar_file.addfile(tar_info, temp_file)
    temp_file.close()


def add_items_to_tarfile_custom(tar_file, items):
    """ Add items assuming they have a custom write function """
    for key, value in items.items():
        temp_file = tempfile.TemporaryFile()
        value.write(temp_file)
        temp_file.seek(0)
        tar_info = tar_file.gettarinfo(arcname=key, fileobj=temp_file)
        tar_file.addfile(tar_info, temp_file)
        temp_file.close()


def add_list_to_tarfile_custom(tar_file, objects, arcname):
    """ Add list of objects assuming they have a custom write function """
    for i, obj in enumerate(objects):
        temp_file = tempfile.TemporaryFile()
        obj.write(temp_file)
        temp_file.seek(0)
        fname = '{}_{}'.format(arcname, i)
        tar_info = tar_file.gettarinfo(arcname=fname, fileobj=temp_file)
        tar_file.addfile(tar_info, temp_file)
        temp_file.close()


def read_items_hdf5(tar_file, arcname):
    """ Read items from hdf5file inside tar_file """

    # read hdf5
    temp_file = tempfile.NamedTemporaryFile()
    temp_file.write(tar_file.extractfile(arcname).read())
    temp_file.seek(0)
    hf = h5py.File(temp_file.name, 'r')
    items = {key: value[:] for key, value in hf.items()}
    hf.close()
    return items


def read_items_pickle(tar_file, arcname):
    items = dict()
    items = pickle.load(tar_file.extractfile(arcname))
    return items


def read_list_custom(tar_file, arcname, read_function, **kwargs):
    objects = []
    i = 0
    while True:
        try:
            fname = '{}_{}'.format(arcname, i)
            f = tar_file.extractfile(fname)
            obj = read_function(f, **kwargs)
            objects.append(obj)
            f.close()
        except KeyError:
            break
        i += 1
    return objects



"""
The ``utilities`` module contains various support functions and classes.
"""
from collections import defaultdict
from ..io.logging import Progress
from sympy.core import S
import sympy
__all__ = ['Progress']


class SparseMatrix(sympy.SparseMatrix):

    def rref_sparse(self, simplify=False):
        """ A sparse row reduce algorithm mimicing the dense version """

        def multiply_and_add_row(row1, factor, row2):
            """ does row1 += factor * row2 when rows are represented by dicts
            """
            keys_to_delete = []
            for k, v in row2.items():
                row1[k] += factor * v
                if row1[k] == 0:
                    keys_to_delete.append(k)
            for k in keys_to_delete:
                del row1[k]

        # The matrix is represented as a list of dicts where each dict is a row
        M = defaultdict(lambda: defaultdict(lambda: 0))

        # Init our special representation. This is possible due to defaultdict
        for r, c, v in self.row_list():
            M[r][c] = v
        M = list(M.values())

        # The pivot elements is stored as tuples as (row, col)
        pivots = []
        r, c = 0, 0  # current row and col of possible pivot
        nRows, nCols = len(M), self.shape[1]
        while r < nRows and c < nCols:
            row = M[r]
            # check if proposed pivot i zero. if so swap this row with a row
            # below which has non zero element at that col
            if c not in row:
                for r2 in range(r + 1, nRows):
                    row2 = M[r2]
                    if c in row2:  # This row has element in the current col
                        M[r], M[r2] = row2, row  # swap the rows
                        row = row2
                        break
                else:  # The pivot and all elements below are zero.
                    c += 1  # goto next column but stay on row
                    continue
            pivots.append((r, c))
            # Normalize row
            row_c = row[c]
            for k in row:
                row[k] /= row_c
            # Start elimination
            r2 = r + 1
            while r2 < nRows:
                row2 = M[r2]
                if c not in row2:
                    r2 += 1
                    continue
                multiply_and_add_row(row2, -row2[c], row)
                if len(row2) == 0:
                    nRows -= 1
                    del M[r2]
                    continue
                r2 += 1
            r += 1
            c += 1

        # Eliminate elements above pivots
        for (r, p) in pivots:
            row_p = M[r]
            for i in range(r):
                row = M[i]
                if p in row:
                    multiply_and_add_row(row, -row[p], row_p)

        # Create the new rrefd matrix
        M2 = SparseMatrix(*self.shape, 0)
        for i, d in enumerate(M):
            for j, v in d.items():
                M2[i, j] = v

        pivots = tuple(p[1] for p in pivots)

        return M2, pivots

    def nullspace(self, simplify=False):
        """ This is a slightly patched version which also uses the sparse rref
        """
        if (max(*self.shape) < 10):  # If matrix small use the dense version
            reduced, pivots = self.rref(simplify=simplify)
        else:
            reduced, pivots = self.rref_sparse(simplify=simplify)

        free_vars = [i for i in range(self.cols) if i not in pivots]

        basis = []
        for free_var in free_vars:
            # for each free variable, we will set it to 1 and all others
            # to 0.  Then, we will use back substitution to solve the system
            vec = [S.Zero]*self.cols
            vec[free_var] = S.One
            for piv_row, piv_col in enumerate(pivots):
                vec[piv_col] -= reduced[piv_row, free_var]
            basis.append(vec)
        return [self._new(self.cols, 1, b) for b in basis]


class BiMap:
    """Simple list like structure with fast dict-lookup

    The structure can append objects and supports some simple list interfaces.
    The lookup is fast since an internal dict stores the indices.
    """
    def __init__(self):
        self._list = list()
        self._dict = dict()

    def __contains__(self, value):
        return value in self._dict

    def __getitem__(self, index):
        return self._list[index]

    def append(self, value):
        """bm.append(hashable) -> None -- append hashable object to end"""
        self._dict[value] = len(self._list)
        self._list.append(value)

    def __len__(self):
        return len(self._list)

    def index(self, value):
        """bm.index(values) -> integer -- return index of value
        Raises ValueError if the values is not present.
        """
        try:
            return self._dict[value]
        except KeyError:
            print(self._dict)
            raise ValueError('{} is not in list'.format(value))

    def copy(self):
        """bm.copy() -> BiMap -- a shallow copy of bm"""
        tbm = BiMap()
        tbm._list = self._list.copy()
        tbm._dict = self._dict.copy()
        return tbm



"""
Ensemble Optimizer

https://en.wikipedia.org/wiki/Bootstrap_aggregating
http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.BaggingRegressor.html  # NOQA
"""

import numpy as np
from typing import Any, Dict, List, Tuple, Union
from .base_optimizer import BaseOptimizer
from .optimizer import Optimizer


class EnsembleOptimizer(BaseOptimizer):
    """
    The ensemble optimizer carries out a series of single optimization runs
    using the :class:`Optimizer` class in order to solve the linear
    :math:`\\boldsymbol{A}\\boldsymbol{x} = \\boldsymbol{y}` problem.
    Subsequently, it provides access to various ensemble averaged
    quantities such as errors and parameters.

    Warning
    -------
    Repeatedly setting up a EnsembleOptimizer and training
    *without* changing the seed for the random number generator will yield
    identical or correlated results, to avoid this please specify a different
    seed when setting up multiple EnsembleOptimizer instances.

    Parameters
    ----------
    fit_data : tuple(numpy.ndarray, numpy.ndarray)
        the first element of the tuple represents the fit matrix `A`
        (`N, M` array) while the second element represents the vector
        of target values `y` (`N` array); here `N` (=rows of `A`,
        elements of `y`) equals the number of target values and `M`
        (=columns of `A`) equals the number of parameters
    fit_method : str
        method to be used for training; possible choice are
        "least-squares", "lasso", "elasticnet", "bayesian-ridge", "ardr",
        "rfe", "split-bregman"
    standardize : bool
        if True the fit matrix and target values are standardized before fitting,
        meaning columns in the fit matrix and th target values are rescaled to
        have a standard deviation of 1.0.
    ensemble_size : int
        number of fits in the ensemble
    train_size : float or int
        if float represents the fraction of `fit_data` (rows) to be used for
        training; if int, represents the absolute number of rows to be used for
        training
    bootstrap : bool
        if True sampling will be carried out with replacement
    check_condition : bool
        if True the condition number will be checked
        (this can be sligthly more time consuming for larger
        matrices)
    seed : int
        seed for pseudo random number generator
    """

    def __init__(self,
                 fit_data: Tuple[np.ndarray, np.ndarray],
                 fit_method: str = 'least-squares',
                 standardize: bool = True,
                 ensemble_size: int = 50,
                 train_size: Union[int, float] = 1.0,
                 bootstrap: bool = True,
                 check_condition: bool = True,
                 seed: int = 42,
                 **kwargs) -> None:

        super().__init__(fit_data, fit_method, standardize, check_condition,
                         seed)

        # set training size
        if isinstance(train_size, float):
            self._train_size = int(np.round(train_size * self.n_target_values))
        elif isinstance(train_size, int):
            self._train_size = train_size
        else:
            raise TypeError('Training size must be int or float')

        self._ensemble_size = ensemble_size
        self._bootstrap = bootstrap
        self._kwargs = kwargs
        self._train_set_list = None
        self._test_set_list = None
        self._parameter_vectors = None
        self._parameters_std = None
        self._rmse_train_ensemble = None
        self._rmse_test_ensemble = None

    def train(self) -> None:
        """
        Carries out ensemble training and construct the final model by
        averaging over all models in the ensemble.
        """
        self._run_ensemble()
        self._construct_final_model()

    def _run_ensemble(self) -> None:
        """ Constructs an ensemble of models. """

        rs = np.random.RandomState(self.seed)
        optimizers = []
        for _ in range(self.ensemble_size):
            # construct training and test sets
            train_set = rs.choice(np.arange(self.n_target_values),
                                  self.train_size, replace=self.bootstrap)
            test_set = np.setdiff1d(
                range(self.n_target_values), train_set)

            # train
            opt = Optimizer((self._A, self._y), self.fit_method,
                            standardize=self.standardize,
                            train_set=train_set, test_set=test_set,
                            check_condition=self._check_condition,
                            **self._kwargs)
            opt.train()
            optimizers.append(opt)

        # collect data from each fit

        self._parameter_vectors = np.array(
            [opt.parameters for opt in optimizers])
        self._train_set_list = [opt.train_set for opt in optimizers]
        self._test_set_list = [opt.test_set for opt in optimizers]
        self._rmse_train_ensemble = np.array(
            [opt.rmse_train for opt in optimizers])
        self._rmse_test_ensemble = np.array(
            [opt.rmse_test for opt in optimizers])

    def _construct_final_model(self) -> None:
        """
        Constructs final model by averaging over all models in the ensemble.
        """
        self._fit_results['parameters'] = np.mean(
            self.parameter_vectors, axis=0)
        self._parameters_std = np.std(self.parameter_vectors, axis=0)

    def predict(self,
                A: np.ndarray,
                return_std: bool = False) \
            -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """
        Predicts data given an input matrix :math:`\boldsymbol{A}`,
        i.e., :math:`\\boldsymbol{A}\\boldsymbol{x}`, where
        :math:`\\boldsymbol{x}` is the vector of the fitted parameters.
        The method returns the vector of predicted values and optionally also
        the vector of standard deviations.

        By using all parameter vectors in the ensemble a standard deviation of
        the prediction can be obtained.

        Parameters
        ----------
        A
            fit matrix where `N` (=rows of `A`, elements of `y`) equals the
            number of target values and `M` (=columns of `A`) equals the number
            of parameters
        return_std
            whether or not to return the standard deviation of the prediction
        """
        prediction = np.dot(A, self.parameters)
        if return_std:
            predictions = np.dot(A, self.parameter_vectors.T)
            if len(predictions.shape) == 1:  # shape is (N, )
                std = np.std(predictions)
            else:  # shape is (N, M)
                std = np.std(predictions, axis=1)
            return prediction, std
        else:
            return prediction

    @property
    def error_matrix(self) -> np.ndarray:
        """
        matrix of fit errors where `N` is the number of target values and
        `M` is the number of fits (i.e., the size of the ensemble)
        """
        if self.parameter_vectors is None:
            return None
        error_matrix = np.zeros((self._n_rows, self.ensemble_size))
        for i, parameters in enumerate(self.parameter_vectors):
            error_matrix[:, i] = np.dot(self._A, parameters) - self._y
        return error_matrix

    @property
    def summary(self) -> Dict[str, Any]:
        """ comprehensive information about the optimizer """
        info = super().summary

        # Add class specific data
        info['parameters_std'] = self.parameters_std
        info['ensemble_size'] = self.ensemble_size
        info['rmse_train'] = self.rmse_train
        info['rmse_train_ensemble'] = self.rmse_train_ensemble
        info['rmse_test'] = self.rmse_test
        info['rmse_test_ensemble'] = self.rmse_test_ensemble
        info['train_size'] = self.train_size
        info['bootstrap'] = self.bootstrap

        # add kwargs used for fitting
        info = {**info, **self._kwargs}
        return info

    def __repr__(self) -> str:
        kwargs = dict()
        kwargs['fit_method'] = self.fit_method
        kwargs['ensemble_size'] = self.ensemble_size
        kwargs['train_size'] = self.train_size
        kwargs['bootstrap'] = self.bootstrap
        kwargs['seed'] = self.seed
        kwargs = {**kwargs, **self._kwargs}
        return 'EnsembleOptimizer((A, y), {})'.format(
            ', '.join('{}={}'.format(*kwarg) for kwarg in kwargs.items()))

    @property
    def parameters_std(self) -> np.ndarray:
        """ standard deviation for each parameter """
        return self._parameters_std

    @property
    def parameter_vectors(self) -> List[np.ndarray]:
        """ all parameter vectors in the ensemble """
        return self._parameter_vectors

    @property
    def ensemble_size(self) -> int:
        """ number of train rounds """
        return self._ensemble_size

    @property
    def rmse_train(self) -> float:
        """
        ensemble average of root mean squared error over train sets
        """
        if self.rmse_train_ensemble is None:
            return None
        return np.sqrt(np.mean((self.rmse_train_ensemble)**2))

    @property
    def rmse_train_ensemble(self) -> np.ndarray:
        """
        root mean squared train errors obtained during for
        each fit in ensemble
        """
        return self._rmse_train_ensemble

    @property
    def rmse_test(self) -> float:
        """
        ensemble average of root mean squared error over test sets
        """
        if self.rmse_test_ensemble is None:
            return None
        return np.sqrt(np.mean((self.rmse_test_ensemble)**2))

    @property
    def rmse_test_ensemble(self) -> np.ndarray:
        """
        root mean squared test errors obtained during for
        each fit in ensemble
        """
        return self._rmse_test_ensemble

    @property
    def train_size(self) -> int:
        """
        number of rows included in train sets; note that this will
        be different from the number of unique rows if boostrapping
        """
        return self._train_size

    @property
    def train_fraction(self) -> float:
        """
        fraction of input data used for training; this value can differ
        slightly from the value set during initialization due to
        rounding
        """
        return self.train_set_size / self._n_rows

    @property
    def bootstrap(self) -> bool:
        """ True if sampling is carried out with replacement """
        return self._bootstrap



"""
Module for generating rattled structures. Rattle refers to displacing atoms
with a normal distribution with zero mean and some standard deviation.
"""

import numpy as np
from scipy.special import erf
from ase.neighborlist import NeighborList


def generate_rattled_structures(atoms, n_structures, rattle_std, seed=42):
    """Returns list of rattled configurations.

    Displacements are drawn from normal distributions for each
    Cartesian directions for each atom independently.

    Warning
    -------
    Repeatedly calling this function *without* providing different
    seeds will yield identical or correlated results. To avoid this
    behavior it is recommended to specify a different seed for each
    call to this function.

    Parameters
    ----------
    atoms : ase.Atoms
        prototype structure
    n_structures : int
        number of structures to generate
    rattle_std : float
        rattle amplitude (standard deviation of the normal distribution)
    seed : int
        seed for setting up NumPy random state from which random numbers are
        generated

    Returns
    -------
    list of ase.Atoms
        generated structures
    """
    rs = np.random.RandomState(seed)
    N = len(atoms)
    atoms_list = []
    for _ in range(n_structures):
        atoms_tmp = atoms.copy()
        displacements = rs.normal(0.0, rattle_std, (N, 3))
        atoms_tmp.positions += displacements
        atoms_list.append(atoms_tmp)
    return atoms_list


def generate_mc_rattled_structures(atoms, n_configs, rattle_std, d_min,
                                   seed=42, **kwargs):
    """Returns list of Monte Carlo rattled configurations.

    Rattling atom `i` is carried out as a Monte Carlo move that is
    accepted with a probability determined from the minimum
    interatomic distance :math:`d_{ij}`.  If :math`\\min(d_{ij})` is
    smaller than :math:`d_{min}` the move is only accepted with a low
    probability.

    This process is repeated for each atom a number of times meaning
    the magnitude of the final displacements is not *directly*
    connected to `rattle_std`.

    Warning
    -------
    Repeatedly calling this function *without* providing different
    seeds will yield identical or correlated results. To avoid this
    behavior it is recommended to specify a different seed for each
    call to this function.

    Notes
    ------
    The procedure implemented here might not generate a symmetric
    distribution for the displacements `kwargs` will be forwarded to
    `mc_rattle` (see user guide for a detailed explanation)

    Parameters
    ----------
    atoms : ase.Atoms
        prototype structure
    n_structures : int
        number of structures to generate
    rattle_std : float
        rattle amplitude (standard deviation in normal distribution);
        note this value is not *directly* connected to the final
        average displacement for the structures
    d_min : float
        interatomic distance used for computing the probability for each rattle
        move
    seed : int
        seed for setting up NumPy random state from which random numbers are
        generated

    Returns
    -------
    list of ase.Atoms
        generated structures
    """
    rs = np.random.RandomState(seed)
    atoms_list = []
    for _ in range(n_configs):
        atoms_tmp = atoms.copy()
        seed = rs.randint(1, 1000000000)
        displacements = mc_rattle(atoms_tmp, rattle_std, d_min, seed=seed,
                                  **kwargs)
        atoms_tmp.positions += displacements
        atoms_list.append(atoms_tmp)
    return atoms_list


def _probability_mc_rattle(d, d_min, width):
    """ Monte Carlo probability function as an error function.

    Parameters
    ----------
    d_min : float
        center value for the error function
    width : float
        width of error function
    """

    return (erf((d-d_min)/width) + 1.0) / 2


def mc_rattle(atoms, rattle_std, d_min, width=0.1, n_iter=10,
              max_attempts=5000, max_disp=2.0, active_atoms=None, seed=42):
    """Generate displacements using the Monte Carlo rattle method

    Parameters
    ----------
    atoms : ase.Atoms
        prototype structure
    rattle_std : float
        rattle amplitude (standard deviation in normal distribution)
    d_min : float
        interatomic distance used for computing the probability for each rattle
        move. Center position of the error function
    width : float
        width of the error function
    n_iter : int
        number of Monte Carlo cycle
    max_disp : float
        rattle moves that yields a displacement larger than max_disp will
        always be rejected. This rarley occurs and is more used as a safety net
        for not generating structures where two or more have swapped positions.
    max_attempts : int
        limit for how many attempted rattle moves are allowed a single atom;
        if this limit is reached an `Exception` is raised.
    active_atoms : list
        list of which atomic indices should undergo Monte Carlo rattling
    seed : int
        seed for setting up NumPy random state from which random numbers are
        generated

    Returns
    -------
    numpy.ndarray
        atomic displacements (`Nx3`)
    """
    rs = np.random.RandomState(seed)

    if active_atoms is None:
        active_atoms = range(len(atoms))

    atoms_rattle = atoms.copy()
    reference_positions = atoms_rattle.get_positions()
    nbr_list = NeighborList([d_min]*len(atoms_rattle), skin=0.0,
                            self_interaction=False, bothways=True)
    nbr_list.update(atoms_rattle)

    # run Monte Carlo
    for _ in range(n_iter):
        for i in active_atoms:
            i_nbrs = np.setdiff1d(nbr_list.get_neighbors(i)[0], [i])
            for n in range(max_attempts):
                delta_disp = rs.normal(0.0, rattle_std, 3)
                atoms_rattle.positions[i] += delta_disp
                disp_i = atoms_rattle.positions[i] - reference_positions[i]
                if np.linalg.norm(disp_i) > max_disp:
                    continue
                min_distance = np.min(atoms_rattle.get_distances(i, i_nbrs, mic=True))
                if _probability_mc_rattle(min_distance, d_min, width) > rs.rand():  # accept disp_i
                    break
                else:  # revert disp_i
                    atoms_rattle[i].position -= delta_disp
            else:
                raise Exception('Maxmium attempts for atom {}'.format(i))
    displacements = atoms_rattle.positions - reference_positions
    return displacements



"""
This module provides functions for reading and writing data files
in phonopy and phono3py formats.
"""

import warnings
import os
import numpy as np
from itertools import product
import hiphive
from ..io.logging import logger

with warnings.catch_warnings():
    warnings.filterwarnings('ignore', category=FutureWarning)
    import h5py


logger = logger.getChild('io_phonopy')


def _filename_to_format(filename):
    """ Tries to guess the format from the filename """
    basename = os.path.basename(filename)

    if basename == 'FORCE_CONSTANTS':
        return 'text'

    if '.' in basename:
        extension = basename.split('.')[-1]
        if extension == 'hdf5':
            return 'hdf5'
    raise ValueError('Could not guess file format')


def read_phonopy_fc2(filename, format=None):
    """Parses a second-order force constant file in phonopy format.

    Parameters
    ----------
    filename : str
        input file name
    format : str
       specify the file-format, if None tries to guess the format from filename

    Returns
    -------
    numpy.ndarray
        second-order force constant matrix (`N,N,3,3` array)
        where `N` is the number of atoms
    """
    fc2_readers = {
        'text': _read_phonopy_fc2_text,
        'hdf5': _read_phonopy_fc2_hdf5
    }

    if format is None:
        format = _filename_to_format(filename)
    if format not in fc2_readers.keys():
        raise ValueError('Did not recognize format {}'.format(format))
    return fc2_readers[format](filename)


def write_phonopy_fc2(filename, fc2, format=None):
    """Writes second-order force constant matrix in phonopy format.

    Parameters
    ----------
    filename : str
        output file name
    fc2 : ForceConstants or numpy.ndarray
        second-order force constant matrix
    format : str
       specify the file-format, if None try to guess the format from filename
     """

    fc2_writers = {
        'text': _write_phonopy_fc2_text,
        'hdf5': _write_phonopy_fc2_hdf5
    }

    # get fc2_array
    if isinstance(fc2, hiphive.force_constants.ForceConstants):
        fc2_array = fc2.get_fc_array(order=2)
    elif isinstance(fc2, np.ndarray):
        fc2_array = fc2
    else:
        raise TypeError('fc2 should be ForceConstants or NumPy array')

    # check that fc2 has correct shape
    n_atoms = fc2_array.shape[0]
    if fc2_array.shape != (n_atoms, n_atoms, 3, 3):
        raise ValueError('fc2 has wrong shape')

    # write
    if format is None:
        format = _filename_to_format(filename)
    if format not in fc2_writers.keys():
        raise ValueError('Did not recognize format {}'.format(format))
    fc2_writers[format](filename, fc2_array)


def read_phonopy_fc3(filename):
    """Parses a third order force constant file in phonopy hdf5 format.

    Parameters
    ----------
    filename : str
        input file name

    Returns
    -------
    numpy.ndarray
        third order force constant matrix
    """
    with h5py.File(filename, 'r') as hf:
        if 'fc3' in hf.keys():
            fc3 = hf['fc3'][:]
        else:
            raise IOError('Could not find fc3 in file {}'.format(filename))
    return fc3


def write_phonopy_fc3(filename, fc3):
    """Writes third order force constant matrix in phonopy hdf5 format.

    Parameters
    ----------
    filename : str
        output file name
    fc3 : ForceConstants or numpy.ndarray
        third order force constant matrix
    """

    if isinstance(fc3, hiphive.force_constants.ForceConstants):
        fc3_array = fc3.get_fc_array(order=3)
    elif isinstance(fc3, np.ndarray):
        fc3_array = fc3
    else:
        raise TypeError('fc3 should be ForceConstants or NumPy array')

    # check that fc3 has correct shape
    n_atoms = fc3_array.shape[0]
    if fc3_array.shape != (n_atoms, n_atoms, n_atoms, 3, 3, 3):
        raise ValueError('fc3 has wrong shape')

    with h5py.File(filename, 'w') as hf:
        hf.create_dataset('fc3', data=fc3_array, compression='gzip')
        hf.flush()


def _read_phonopy_fc2_text(filename):
    """ Reads phonopy-fc2 file in text format. """
    with open(filename, 'r') as f:

        # read shape of force constants
        line = f.readline()
        line_ints = [int(x) for x in line.split()]
        if len(line_ints) == 1:
            n_atoms = line_ints[0]
        elif len(line_ints) == 2:
            assert line_ints[0] == line_ints[1]
            n_atoms = line_ints[0]
        else:
            raise ValueError('Unsupported or incorrect phonopy format')
        fc2 = np.full((n_atoms, n_atoms, 3, 3), np.nan)

        # read force constants
        lines = f.readlines()
        for n, line in enumerate(lines):
            flds = line.split()
            if len(flds) == 2:
                i = int(flds[0]) - 1  # phonopy index starts with 1
                j = int(flds[1]) - 1
                for x in range(3):
                    fc_row = lines[n + x + 1].split()
                    for y in range(3):
                        fc2[i][j][x][y] = float(fc_row[y])
    return fc2


def _read_phonopy_fc2_hdf5(filename):
    """ Reads phonopy-fc2 file in hdf5 format. """
    with h5py.File(filename, 'r') as hf:
        if 'force_constants' in hf.keys():
            fc2 = hf['force_constants'][:]
        elif 'fc2' in hf.keys():
            fc2 = hf['fc2'][:]
        else:
            raise IOError('Could not find fc2 in file {}'.format(filename))
    return fc2


def _write_phonopy_fc2_text(filename, fc2):
    """ Writes second-order force constants to file in text format. """
    n_atoms = fc2.shape[0]
    with open(filename, 'w') as f:
        f.write('{:5d} {:5d}\n'.format(n_atoms, n_atoms))
        for i, j in product(range(n_atoms), repeat=2):
            fc2_ij = fc2[(i, j)]
            if fc2_ij.shape != (3, 3):
                raise ValueError('Invalid shape for fc2[({},{})]'.format(i, j))
            f.write('{:-5d}{:5d}\n'.format(i + 1, j + 1))
            for row in fc2_ij:
                f.write((3*' {:22.15f}'+'\n').format(*tuple(row)))


def _write_phonopy_fc2_hdf5(filename, fc2):
    """ Writes second-order force constants to file in hdf5 format. """
    with h5py.File(filename, 'w') as hf:
        hf.create_dataset('fc2', data=fc2, compression='gzip')
        hf.flush()



"""
scikit-learn is an excellent library for training linear models and provides a
large number of useful tools.

This module provides simplified interfaces for vaiours linear model regression
methods. These methods are set up in a way that work out of the box for typical
problems in cluster expansion and force constant potential construction. This
includes slight adjustments scitkit-learn default values.

If you would like more flexibility or extended functionality or ability to
fine-tune parameters that are not included in this interface, it is of course
possible to use scikit-learn directly.
More information about the sklearn linear models can be found at
http://scikit-learn.org/stable/modules/linear_model.html
"""

import numpy as np
from collections import OrderedDict
from sklearn.linear_model import (Lasso,
                                  LassoCV,
                                  Ridge,
                                  RidgeCV,
                                  ElasticNet,
                                  ElasticNetCV,
                                  BayesianRidge,
                                  ARDRegression)
from sklearn.model_selection import ShuffleSplit
from sklearn.feature_selection import RFE, RFECV
from sklearn.preprocessing import StandardScaler
from typing import Any, Dict, List, Union
from ..io.logging import logger
from .split_bregman import fit_split_bregman


logger = logger.getChild('fit_methods')


def fit(X: np.ndarray,
        y: np.ndarray,
        fit_method: str,
        standardize: bool = True,
        check_condition: bool = True,
        **kwargs) -> Dict[str, Any]:
    """
    Wrapper function for all available fit methods.  The function
    returns parameters and other pertinent information in the form of
    a dictionary.

    Parameters
    -----------
    X
        fit matrix
    y
        target array
    fit_method
        method to be used for training; possible choice are
        "least-squares", "lasso", "elasticnet", "bayesian-ridge", "ardr",
        "rfe", "split-bregman"
    standardize : bool
        if True the fit matrix is standardized before fitting
    check_condition : bool
        if True the condition number will be checked
        (this can be sligthly more time consuming for larger
        matrices)
    """

    if fit_method not in available_fit_methods:
        msg = ['Fit method not available']
        msg += ['Please choose one of the following:']
        for key in available_fit_methods:
            msg += [' * ' + key]
        raise ValueError('\n'.join(msg))

    if check_condition:
        cond = np.linalg.cond(X)
        if cond > 1e10:
            logger.warning('Condition number is large, {}'.format(cond))

    if standardize:

        # standardize fit matrix, column wise std -> 1.0
        ss = StandardScaler(copy=False, with_mean=False, with_std=True)
        ss.fit_transform(X)  # change in place

        # standardize target values, std(y) -> 1.0
        y_scale = 1.0/np.std(y)
        y_rescaled = y * y_scale

        # train
        results = fit_methods[fit_method](X, y_rescaled, **kwargs)

        # inverse standardization
        parameters = results['parameters'] / y_scale
        ss.inverse_transform(X)  # change in place
        ss.transform(parameters.reshape(1, -1)).reshape(-1,)
        results['parameters'] = parameters

    else:
        results = fit_methods[fit_method](X, y, **kwargs)
    return results


def _fit_least_squares(X: np.ndarray, y: np.ndarray) -> Dict[str, Any]:
    """
    Returns the least-squares solution `a` to the linear problem
    `Xa=y` in the form of a dictionary with a key named `parameters`.

    This function is a wrapper to the `linalg.lstsq` function in NumPy.

    Parameters
    -----------
    X
        fit matrix
    y
        target array
    """
    results = dict()
    results['parameters'] = np.linalg.lstsq(X, y, rcond=-1)[0]
    return results


def _fit_lasso(X: np.ndarray, y: np.ndarray,
               alpha: float = None, fit_intercept: bool = False,
               **kwargs) -> Dict[str, Any]:
    """
    Returns the solution `a` to the linear problem `Xa=y` obtained by
    using the LASSO method as implemented in scitkit-learn in the form
    of a dictionary with a key named `parameters`.

    LASSO optimizes the following problem::

        (1 / (2 * n_samples)) * ||y - Xw||^2_2 + alpha * ||w||_1

    If `alpha` is `None` this function will call `fit_lassoCV` which attempts
    to find the optimal alpha via sklearn's `LassoCV` class.

    Parameters
    ----------
    X
        fit matrix
    y
        target array
    alpha
        alpha value
    fit_intercept
        center data or not, forwarded to sklearn
    """
    if alpha is None:
        return _fit_lassoCV(X, y, fit_intercept=fit_intercept, **kwargs)
    else:
        lasso = Lasso(alpha=alpha, fit_intercept=fit_intercept, **kwargs)
        lasso.fit(X, y)
        results = dict()
        results['parameters'] = lasso.coef_
        return results


def _fit_lassoCV(X: np.ndarray,
                 y: np.ndarray,
                 alphas: List[float] = None,
                 fit_intercept: bool = False,
                 cv: int = 10,
                 n_jobs: int = -1,
                 **kwargs) -> Dict[str, Any]:
    """
    Returns the solution `a` to the linear problem `Xa=y` obtained by
    using the LassoCV method as implemented in scitkit-learn in the
    form of a dictionary with a key named `parameters`.

    The dictionary will also contain the keys `alpha_optimal` (alpha
    value that yields the lowest validation RMSE), `alpha_path` (all
    tested alpha values), and `mse_path` (MSE for validation set for
    each alpha).

    Parameters
    -----------
    X
        fit matrix
    y
        target array
    alphas
        list of alpha values to be evaluated during regularization path
    fit_intercept
        center data or not, forwarded to sklearn
    cv
        how many folds to carry out in cross-validation
    n_jobs
        number of cores to use during the cross validation.
        None means 1 unless in a joblib.parallel_backend context.
        -1 means using all processors.
        See sklearn's glossary for more details.
    """
    if alphas is None:
        alphas = np.logspace(-8, -0.3, 100)

    lassoCV = LassoCV(alphas=alphas, fit_intercept=fit_intercept, cv=cv, n_jobs=n_jobs, **kwargs)
    lassoCV.fit(X, y)
    results = dict()
    results['parameters'] = lassoCV.coef_
    results['alpha_optimal'] = lassoCV.alpha_
    return results


def _fit_ridge(X, y, alpha=None, fit_intercept=False, **kwargs):
    results = dict()
    if alpha is None:
        if 'alphas' not in kwargs:
            kwargs['alphas'] = np.logspace(-6, 3, 100)
        ridge = RidgeCV(fit_intercept=fit_intercept, **kwargs)
        ridge.fit(X, y)
        results['alpha_optimal'] = ridge.alpha_
    else:
        ridge = Ridge(alpha=alpha, fit_intercept=fit_intercept, **kwargs)
        ridge.fit(X, y)
    results['parameters'] = ridge.coef_
    return results


def _fit_bayesian_ridge(X: np.ndarray, y: np.ndarray,
                        fit_intercept: bool = False,
                        **kwargs) -> Dict[str, Any]:
    """
    Returns the solution `a` to the linear problem `Xa=y` obtained by using
    Bayesian ridge regression as implemented in scitkit-learn in the
    form of a dictionary with a key named `parameters`.

    Parameters
    -----------
    X
        fit matrix
    y
        target array
    fit_intercept
        center data or not, forwarded to sklearn
    """
    brr = BayesianRidge(fit_intercept=fit_intercept, **kwargs)
    brr.fit(X, y)
    results = dict()
    results['parameters'] = brr.coef_
    return results


def _fit_elasticnet(X: np.ndarray, y: np.ndarray,
                    alpha: float = None, fit_intercept: bool = False,
                    **kwargs) -> Dict[str, Any]:
    """
    Returns the solution `a` to the linear problem `Xa=y` obtained by using
    the ElasticNet method as implemented in scitkit-learn in the
    form of a dictionary with a key named `parameters`.

    If `alpha` is `None` this function will call the fit_lassoCV which attempts
    to find the optimal alpha via sklearn ElasticNetCV class.

    Parameters
    -----------
    X
        fit matrix
    y
        target array
    alpha
        alpha value
    fit_intercept
        center data or not, forwarded to sklearn
    """
    if alpha is None:
        return _fit_elasticnetCV(X, y, fit_intercept=fit_intercept, **kwargs)
    else:
        elasticnet = ElasticNet(alpha=alpha, fit_intercept=fit_intercept, **kwargs)
        elasticnet.fit(X, y)
        results = dict()
        results['parameters'] = elasticnet.coef_
        return results


def _fit_elasticnetCV(X: np.ndarray,
                      y: np.ndarray,
                      alphas: List[float] = None,
                      l1_ratio: Union[float, List[float]] = None,
                      fit_intercept: bool = False,
                      cv: int = 10,
                      n_jobs: int = -1,
                      **kwargs) -> Dict[str, Any]:
    """
    Returns the solution `a` to the linear problem `Xa=y` obtained by using
    the ElasticNetCV method as implemented in scitkit-learn in the
    form of a dictionary with a key named `parameters`.

    The dictionary returned by this function will also contain the
    fields `alpha_optimal` (alpha value that yields the lowest
    validation RMSE), `alpha_path` (all tested alpha values),
    `l1_ratio_optmal` (alpha value that yields the lowest validation
    RMSE), `l1_ratio_path` (all tested `l1_ratio` values) `mse_path`
    (MSE for validation set for each alpha and `l1_ratio`)

    Parameters
    -----------
    X
        fit matrix
    y
        target array
    alphas
        list of alpha values to be evaluated during regularization path
    l1_ratio
        l1_ratio values to be evaluated during regularization path
    fit_intercept
        center data or not, forwarded to sklearn
    cv
        how many folds to carry out in cross-validation
    n_jobs
        number of cores to use during the cross validation.
        None means 1 unless in a joblib.parallel_backend context.
        -1 means using all processors.
        See sklearn's glossary for more details.
    """

    if alphas is None:
        alphas = np.logspace(-8, -0.3, 100)
    if l1_ratio is None:
        l1_ratio = [1.0, 0.995, 0.99, 0.98, 0.97, 0.95, 0.925, 0.9, 0.85,
                    0.8, 0.75, 0.65, 0.5, 0.4, 0.25, 0.1]

    elasticnetCV = ElasticNetCV(alphas=alphas, l1_ratio=l1_ratio, cv=cv,
                                fit_intercept=fit_intercept, n_jobs=n_jobs, **kwargs)
    elasticnetCV.fit(X, y)
    results = dict()
    results['parameters'] = elasticnetCV.coef_
    results['alpha_optimal'] = elasticnetCV.alpha_
    results['l1_ratio_optimal'] = elasticnetCV.l1_ratio_
    return results


def _fit_ardr(X: np.ndarray,
              y: np.ndarray,
              threshold_lambda: float = None,
              line_scan: bool = False,
              fit_intercept: bool = False,
              **kwargs) -> Dict[str, Any]:
    """
    Returns the solution `a` to the linear problem `Xa=y` obtained by
    using the automatic relevance determination regression (ARDR)
    method as implemented in scitkit-learn in the form of a dictionary
    with a key named `parameters`.

    Parameters
    -----------
    X
        fit matrix
    y
        target array
    threshold_lambda
        threshold lambda parameter forwarded to sklearn
    line_scan
        whether or not to perform line-scan in order to find optimal
        threshold-lambda
    fit_intercept
        center data or not, forwarded to sklearn
    """

    if threshold_lambda is not None and line_scan:
        raise ValueError('Specify threshold_lambda or set line_scan=True, not both')

    if threshold_lambda is None:
        threshold_lambda = 1e4

    if line_scan:
        return _fit_ardr_line_scan(X, y, fit_intercept=fit_intercept, **kwargs)
    else:
        ardr = ARDRegression(threshold_lambda=threshold_lambda, fit_intercept=fit_intercept,
                             **kwargs)
        ardr.fit(X, y)
        results = dict()
        results['parameters'] = ardr.coef_
        return results


def _fit_ardr_line_scan(X: np.ndarray,
                        y: np.ndarray,
                        cv_splits: int = 1,
                        threshold_lambdas: List[float] = None,
                        **kwargs) -> Dict[str, Any]:
    """ ARDR with line-scan for optimal threshold-lambda.

    Parameters
    -----------
    X
        fit matrix
    y
        target array
    threshold_lambdas
        list of threshold-lambda values to be evaluated. The optimal
        lambda-value found will be used in the final fit.
    cv_splits
        how many CV splits to carry out when evaluating each lambda value.
    """

    from .cross_validation import CrossValidationEstimator

    # default lambda values to scan
    if threshold_lambdas is None:
        threshold_lambdas = np.logspace(3, 6, 15)

    # run lin-scan of lambda values
    cv_data = []
    for lam in threshold_lambdas:
        cve = CrossValidationEstimator((X, y), fit_method='ardr', validation_method='shuffle-split',
                                       threshold_lambda=lam, test_size=0.1, train_size=0.9,
                                       n_splits=cv_splits, **kwargs)
        cve.validate()
        cv_data.append([lam, cve.rmse_validation])

    # select best lambda
    cv_data = np.array(cv_data)
    optimal_ind = cv_data[:, 1].argmin()
    lambda_optimal = cv_data[optimal_ind, 0]

    # final fit with optimal lambda
    results = _fit_ardr(X, y, threshold_lambda=lambda_optimal, **kwargs)
    results['threshold_lambda_optimal'] = lambda_optimal
    return results


class _Estimator:
    """ Estimator class making it possible to use all fit methods
    as estimators in RFE """

    def __init__(self, fit_method, **kwargs):
        if fit_method == 'rfe':
            raise ValueError('recursive infinitum')
        self.fit_method = fit_method
        self.kwargs = kwargs
        self.coef_ = None

    def fit(self, X, y):
        fit_func = fit_methods[self.fit_method]
        results = fit_func(X, y, **self.kwargs)
        self.coef_ = results['parameters']

    def get_params(self, deep=True):
        params = {k: v for k, v in self.kwargs.items()}
        params['fit_method'] = self.fit_method
        return params

    def predict(self, A):
        return np.dot(A, self.coef_)


def fit_rfe(X: np.ndarray,
            y: np.ndarray,
            n_features: int = None,
            step: Union[int, float] = 0.04,
            estimator: str = 'least-squares',
            final_estimator: str = None,
            estimator_kwargs: dict = {},
            final_estimator_kwargs: dict = {},
            cv_splits: int = 5,
            n_jobs: int = -1,
            **rfe_kwargs):
    """
    Returns the solution `a` to the linear problem `Xa=y` obtained by
    recursive feature elimination (RFE).

    Parameters
    -----------
    X
        fit matrix
    y
        target array
    n_features
        number of features to select, if None sklearn.feature_selection.RFECV
        will be used to determine the optimal number of features
    step
        if given as integer then corresponds to number of parameters to
        eliminate in each iteration. If given as a float then corresponds to
        the fraction of parameters to remove each iteration.
    estimator
        fit method during RFE algorithm
    final_estimator
        fit_method to be used in final fit,
        if None will default to whichever estimator is being used
    cv_splits
        number of cv-splits to carry out if finding optimal n_features
    n_jobs
        number of cores to use during the cross validation.
        -1 means using all processors.
    """

    # handle kwargs
    if final_estimator is None:
        final_estimator = estimator
        if len(final_estimator_kwargs) == 0:
            final_estimator_kwargs = estimator_kwargs

    estimator_obj = _Estimator(estimator, **estimator_kwargs)
    if n_features is None:
        if 'scoring' not in rfe_kwargs:
            rfe_kwargs['scoring'] = 'neg_mean_squared_error'
        cv = ShuffleSplit(train_size=0.9, test_size=0.1, n_splits=cv_splits)
        rfe = RFECV(estimator_obj, step=step, cv=cv, n_jobs=n_jobs, **rfe_kwargs)
    else:
        rfe = RFE(estimator_obj, n_features_to_select=n_features, step=step, **rfe_kwargs)

    # Carry out RFE
    rfe.fit(X, y)
    features = rfe.support_
    ranking = rfe.ranking_

    # carry out final fit
    n_params = X.shape[1]
    results = fit(X[:, features], y, fit_method=final_estimator, **final_estimator_kwargs)
    params = np.zeros(n_params)
    params[features] = results['parameters']
    results['parameters'] = params

    # finish up
    results['features'] = features
    results['ranking'] = ranking
    return results


fit_methods = OrderedDict([
    ('least-squares', _fit_least_squares),
    ('lasso', _fit_lasso),
    ('ridge', _fit_ridge),
    ('bayesian-ridge', _fit_bayesian_ridge),
    ('elasticnet', _fit_elasticnet),
    ('split-bregman', fit_split_bregman),
    ('ardr', _fit_ardr),
    ('rfe', fit_rfe),
    ])
available_fit_methods = sorted(fit_methods.keys())



import pickle


def _write_pickle(fname, data):
    """ Write data to pickle file with filename fname """
    with open(fname, 'wb') as handle:
        pickle.dump(data, handle, protocol=pickle.HIGHEST_PROTOCOL)


def _read_pickle(fname):
    """ Read pickle file and return content """
    with open(fname, 'rb') as handle:
        return pickle.load(handle)



"""
Functionality for enforcing translational sum rules
"""

import sympy
from scipy.sparse import coo_matrix
import numpy as np

from ..io.logging import logger
from .utilities import SparseMatrix, BiMap
from .orbits import get_orbits


logger = logger.getChild('sum_rules')


def create_constraint_map(cs):

    logger.debug('Starting constructing constraint matrices')

    # First we need to know the global index of a eigentensor given which
    # orbit it belongs to.
    params = []
    n = 0
    for orbit_index, orbit in enumerate(cs.orbits):
        nParams_in_orbit = len(orbit.eigentensors)
        params.append(list(range(n, n + nParams_in_orbit)))
        n += nParams_in_orbit

    # If we dont want the sum rules an eye matrix is created instead.
    if not cs.acoustic_sum_rules:
        row = list(range(n))
        col = list(range(n))
        data = [1.0]*n
        cs._cvs = coo_matrix((data, (row, col)), shape=(n, n))
        return None

    # The lookup is a fast way to get the eigentensors given a cluster Also the
    # orbit index is bundled to quickly find the correct parameter indices
    lookup = {}
    for orbit_index, orbit in enumerate(cs.orbits):
        for of in orbit.orientation_families:
            for cluster_index, perm_index in zip(of.cluster_indices,
                                                 of.permutation_indices):
                cluster = cs._cluster_list[cluster_index]
                perm = cs._permutations[perm_index]
                lookup[tuple(cluster)] = [et.transpose(perm) for et in
                                          of.eigentensors], orbit_index

    # Init just one row to be able to stack. n is the total number of
    # eigentensors
    M = SparseMatrix(1, n, 0)  # This is the constraint matrix
    for order in cs.cutoffs.orders:
        logger.debug('Order {}: '.format(order))

        # Find which minimal amount of prefixes that is needed
        # Essentially they are all the prototypes of order one less still
        # compatible with the cutoff.
        prefix_list = []
        if order == 2:
            for i in np.unique(cs.wyckoff_sites):
                prefix_list.append((i,))
        else:
            tmp_cluster_set = set()
            for cluster in cs._cluster_list:
                if len(cluster) != order:
                    continue
                for i in range(order):
                    c = tuple(cluster[:i]) + tuple(cluster[i+1:])
                    if c[0] < len(cs._prim):
                        tmp_cluster_set.add(c)
            tmp_cluster_list = BiMap()
            for c in tmp_cluster_set:
                tmp_cluster_list.append(c)
            orbits = get_orbits(tmp_cluster_list, cs._atom_list,
                                cs.rotation_matrices, cs.translation_vectors,
                                cs.permutations, cs._prim, cs.symprec)
            for orb in orbits:
                prefix_list.append(tmp_cluster_list[orb.prototype_index])

        m = SparseMatrix(3**order, M.shape[1], 0)  # intermediate matrix
        # Now loop over the prefix and add the last atom index
        for prefix in prefix_list:
            m *= 0
            cluster = list(prefix) + [None]
            for i in range(len(cs._atom_list)):
                cluster[-1] = i
                tmp = lookup.get(tuple(sorted(cluster)))
                if tmp is None:
                    continue
                ets, orbit_index = tmp
                inv_argsort = np.argsort(np.argsort(cluster))
                ets = [et.transpose(inv_argsort) for et in ets]
                for et, col in zip(ets, params[orbit_index]):
                    et_flat = et.flatten()
                    for j in et_flat.nonzero()[0]:
                        m[j, col] += sympy.nsimplify(et_flat[j])
            M = SparseMatrix.vstack(M, m)
            # only compress if there are very many elements
            if M.nnz() > cs._config['max_number_constraint_elements']:
                M = SparseMatrix.rref(M)[0]

    logger.debug('Finished constructing constraint matrices')

    logger.debug('Starting constructing constraint vectors')
    nullspace = M.nullspace()
    logger.debug('Assembling solutions...')
    row = []
    col = []
    data = []
    for c, vec in enumerate(nullspace):
        for r, _, v in vec.row_list():
            row.append(r)
            col.append(c)
            data.append(np.float64(v))
    # This is the final product, the constraint map (or constraint vectors)
    cs._cvs = coo_matrix((data, (row, col)), shape=(len(vec), len(nullspace)))

    logger.debug('{} degrees of freedom'.format(cs._cvs.shape[1]))
    logger.debug('Finished constructing constraint vectors')



"""
Optimizer
"""
import numpy as np
from sklearn.model_selection import train_test_split
from typing import Any, Dict, List, Tuple, Union
from .base_optimizer import BaseOptimizer
from .fit_methods import fit
from .tools import ScatterData


class Optimizer(BaseOptimizer):
    """
    This optimizer finds a solution to the linear
    :math:`\\boldsymbol{A}\\boldsymbol{x}=\\boldsymbol{y}` problem.

    One has to specify either `train_size`/`test_size` or
    `train_set`/`test_set` If either `train_set` or `test_set` (or both)
    is specified the fractions will be ignored.

    Warning
    -------
    Repeatedly setting up a Optimizer and training
    *without* changing the seed for the random number generator will yield
    identical or correlated results, to avoid this please specify a different
    seed when setting up multiple Optimizer instances.

    Parameters
    ----------
    fit_data : tuple(numpy.ndarray, numpy.ndarray)
        the first element of the tuple represents the fit matrix `A`
        (`N, M` array) while the second element represents the vector
        of target values `y` (`N` array); here `N` (=rows of `A`,
        elements of `y`) equals the number of target values and `M`
        (=columns of `A`) equals the number of parameters
    fit_method : str
        method to be used for training; possible choice are
        "least-squares", "lasso", "elasticnet", "bayesian-ridge", "ardr",
        "rfe", "split-bregman"
    standardize : bool
        if True the fit matrix and target values are standardized before fitting,
        meaning columns in the fit matrix and th target values are rescaled to
        have a standard deviation of 1.0.
    train_size : float or int
        If float represents the fraction of `fit_data` (rows) to be used for
        training. If int, represents the absolute number of rows to be used for
        training.
    test_size : float or int
        If float represents the fraction of `fit_data` (rows) to be used for
        testing. If int, represents the absolute number of rows to be used for
        testing.
    train_set : tuple or list(int)
        indices of rows of `A`/`y` to be used for training
    test_set : tuple or list(int)
        indices of rows of `A`/`y` to be used for testing
    check_condition : bool
        if True the condition number will be checked
        (this can be sligthly more time consuming for larger
        matrices)
    seed : int
        seed for pseudo random number generator

    Attributes
    ----------
    train_scatter_data : ScatterData
        target and predicted value for each row in the training set
    test_scatter_data : ScatterData
        target and predicted value for each row in the test set
    """

    def __init__(self,
                 fit_data: Tuple[np.ndarray, np.ndarray],
                 fit_method: str = 'least-squares',
                 standardize: bool = True,
                 train_size: Union[int, float] = 0.75,
                 test_size: Union[int, float] = None,
                 train_set: Union[Tuple[int], List[int]] = None,
                 test_set: Union[Tuple[int], List[int]] = None,
                 check_condition: bool = True,
                 seed: int = 42,
                 **kwargs) -> None:

        super().__init__(fit_data, fit_method, standardize, check_condition,
                         seed)

        self._kwargs = kwargs

        # setup train and test sets
        self._setup_rows(train_size, test_size, train_set, test_set)

        # will be populate once running train
        self._rmse_train = None
        self._rmse_test = None
        self._contributions_train = None
        self._contributions_test = None
        self.train_scatter_data = None
        self.test_scatter_data = None

    def train(self) -> None:
        """ Carries out training. """

        # select training data
        A_train = self._A[self.train_set, :]
        y_train = self._y[self.train_set]

        # perform training
        self._fit_results = fit(A_train, y_train, self.fit_method,
                                self.standardize, self._check_condition,
                                **self._kwargs)
        self._rmse_train = self.compute_rmse(A_train, y_train)
        self.train_scatter_data = ScatterData(y_train, self.predict(A_train))

        # perform testing
        if self.test_set is not None:
            A_test = self._A[self.test_set, :]
            y_test = self._y[self.test_set]
            self._rmse_test = self.compute_rmse(A_test, y_test)
            self.test_scatter_data = ScatterData(y_test, self.predict(A_test))
        else:
            self._rmse_test = None
            self.test_scatter_data = None

    def _setup_rows(self,
                    train_size: Union[int, float],
                    test_size: Union[int, float],
                    train_set: Union[Tuple[int], List[int]],
                    test_set: Union[Tuple[int], List[int]]) -> None:
        """
        Sets up train and test rows depending on which arguments are
        specified.

        If `train_set` and `test_set` are `None` then `train_size` and
        `test_size` are used.
        """

        if train_set is None and test_set is None:
            train_set, test_set = self._get_rows_via_sizes(
                train_size, test_size)
        else:
            train_set, test_set = self._get_rows_from_indices(
                train_set, test_set)

        if len(train_set) == 0:
            raise ValueError('No training rows selected from fit_data')

        if test_set is not None:  # then check overlap between train and test
            if len(np.intersect1d(train_set, test_set)):
                raise ValueError('Overlap between training and test set')
            if len(test_set) == 0:
                test_set = None

        self._train_set = train_set
        self._test_set = test_set

    def _get_rows_via_sizes(self,
                            train_size: Union[int, float],
                            test_size: Union[int, float]) \
            -> Tuple[List[int], List[int]]:
        """ Returns train and test rows via sizes. """

        # Handle special cases
        if test_size is None and train_size is None:
            raise ValueError('Training and test set sizes are None (empty).')
        elif train_size is None and abs(test_size - 1.0) < 1e-10:
            raise ValueError('Traininig set is empty.')

        elif test_size is None:
            if train_size == self._n_rows or abs(train_size-1.0) < 1e-10:
                train_set = np.arange(self._n_rows)
                test_set = None
                return train_set, test_set

        # split
        train_set, test_set = train_test_split(np.arange(self._n_rows),
                                               train_size=train_size,
                                               test_size=test_size,
                                               random_state=self.seed)

        return train_set, test_set

    def _get_rows_from_indices(self,
                               train_set: Union[Tuple[int], List[int]],
                               test_set: Union[Tuple[int], List[int]]) \
            -> Tuple[np.ndarray, np.ndarray]:
        """ Returns rows via indices. """
        if train_set is None and test_set is None:
            raise ValueError('Training and test sets are None (empty)')
        elif test_set is None:
            test_set = [i for i in range(self._n_rows)
                        if i not in train_set]
        elif train_set is None:
            train_set = [i for i in range(self._n_rows)
                         if i not in test_set]
        return np.array(train_set), np.array(test_set)

    @property
    def summary(self) -> Dict[str, Any]:
        """ comprehensive information about the optimizer """
        info = super().summary

        # Add class specific data
        info['rmse_train'] = self.rmse_train
        info['rmse_test'] = self.rmse_test
        info['train_size'] = self.train_size
        info['train_set'] = self.train_set
        info['test_size'] = self.test_size
        info['test_set'] = self.test_set
        info['train_scatter_data'] = self.train_scatter_data
        info['test_scatter_data'] = self.test_scatter_data

        # add kwargs used for fitting
        info = {**info, **self._kwargs}
        return info

    def __repr__(self) -> str:
        kwargs = dict()
        kwargs['fit_method'] = self.fit_method
        kwargs['traininig_size'] = self.train_size
        kwargs['test_size'] = self.test_size
        kwargs['train_set'] = self.train_set
        kwargs['test_set'] = self.test_set
        kwargs['seed'] = self.seed
        kwargs = {**kwargs, **self._kwargs}
        return 'Optimizer((A, y), {})'.format(
            ', '.join('{}={}'.format(*kwarg) for kwarg in kwargs.items()))

    @property
    def rmse_train(self) -> float:
        """ root mean squared error for training set """
        return self._rmse_train

    @property
    def rmse_test(self) -> float:
        """ root mean squared error for test set """
        return self._rmse_test

    @property
    def contributions_train(self) -> np.ndarray:
        """ average contribution to the predicted values for
        the train set from each parameter """
        return self._contributions_train

    @property
    def contributions_test(self) -> np.ndarray:
        """ average contribution to the predicted values for
        the test set from each parameter """
        return self._contributions_test

    @property
    def train_set(self) -> List[int]:
        """ indices of rows included in the training set """
        return self._train_set

    @property
    def test_set(self) -> List[int]:
        """ indices of rows included in the test set """
        return self._test_set

    @property
    def train_size(self) -> int:
        """ number of rows included in training set """
        return len(self.train_set)

    @property
    def train_fraction(self) -> float:
        """ fraction of rows included in training set """
        return self.train_size / self._n_rows

    @property
    def test_size(self) -> int:
        """ number of rows included in test set """
        if self.test_set is None:
            return 0
        return len(self.test_set)

    @property
    def test_fraction(self) -> float:
        """ fraction of rows included in test set """
        if self.test_set is None:
            return 0.0
        return self.test_size / self._n_rows



"""
Module containing tensor related functions
"""

import numpy as np


_paths = dict()


def rotation_tensor_as_matrix(R, order):
    einsum_input = []
    for i in range(order):
        einsum_input.append(R)
        einsum_input.append([i, order + i])
    R_tensor = np.einsum(*einsum_input)
    R_matrix = R_tensor.reshape((3**order, 3**order))
    return R_matrix


def rotate_tensor_precalc(T, R):
    order = len(T.shape)
    return np.dot(T.ravel(), R).reshape((3,) * order)


def rotate_tensor(T, R, path=None):
    """Equivalent to T_abc... = T_ijk... R_ia R_jb R_kc ... """

    order = len(T.shape)
    if order not in _paths:
        path = ['einsum_path', (0, 1)]
        for i in range(order - 1, 0, -1):
            path.append((0, i))
        _paths[order] = path

    einsum_input = [T, list(range(order))]
    for i in range(order):
        einsum_input.append(R)
        einsum_input.append([i, order + i])
    return np.einsum(*einsum_input, optimize=_paths[order])


def rotation_to_cart_coord(R, cell):
    """Return the rotation matrix in cart coord given a cell metric """
    return np.dot(np.dot(cell.T, R), np.linalg.inv(cell.T))



"""
This module contains various support/utility functions.
"""

from typing import List
import numpy as np

from ase import Atoms
from ase.calculators.calculator import Calculator
from ase.geometry import find_mic
from ase.geometry import get_distances
from hiphive import ClusterSpace, ForceConstants
from hiphive.force_constant_model import ForceConstantModel
from .io.logging import logger


logger = logger.getChild('utilities')


def get_displacements(atoms: Atoms, atoms_ideal: Atoms) -> np.ndarray:
    """Returns the the smallest possible displacements between a
    displaced configuration relative to an ideal (reference)
    configuration.

    Notes
    -----
    * uses :func:`ase.geometry.find_mic`
    * assumes periodic boundary conditions in all directions

    Parameters
    ----------
    atoms
        configuration with displaced atoms
    atoms_ideal
        ideal configuration relative to which displacements are computed
    """
    if not np.array_equal(atoms.numbers, atoms_ideal.numbers):
        raise ValueError('Atomic numbers do not match.')

    displacements = []
    for pos, ideal_pos in zip(atoms.positions, atoms_ideal.positions):
        v_ij = np.array([pos - ideal_pos])
        displacements.append(find_mic(v_ij, atoms_ideal.cell, pbc=True)[0][0])
    return np.array(displacements)


def prepare_structures(structures: List[Atoms], atoms_ideal: Atoms,
                       calc: Calculator) -> None:
    """Prepares a set of structures in the format suitable for a
    :class:`StructureContainer <hiphive.StructureContainer>`.  This
    includes retrieving the forces using the provided calculator,
    resetting the positions to the ideal configuration, and adding
    arrays with forces and displacements.

    Notes
    -----
    Changes the structures in place.

    Parameters
    ----------
    structures
        list of input configurations
    atoms_ideal
        reference configuration relative to which displacements
        are computed
    calc
        ASE calculator used for computing forces
    """
    for atoms in structures:
        atoms.set_calculator(calc)
        forces = atoms.get_forces()
        displacements = get_displacements(atoms, atoms_ideal)
        atoms.positions = atoms_ideal.get_positions()
        atoms.new_array('forces', forces)
        atoms.new_array('displacements', displacements)


def find_permutation(atoms: Atoms, atoms_ref: Atoms) -> List[int]:
    """ Returns the best permutation of atoms for mapping one
    configuration onto another.

    Parameters
    ----------
    atoms
        configuration to be permuted
    atoms_ref
        configuration onto which to map

    Examples
    --------
    After obtaining the permutation via
    ```
    p = find_permutation(atoms1, atoms2)
    ```
    the reordered structure ``atoms1[p]`` will give the closest match
    to ``atoms2``.
    """
    assert np.linalg.norm(atoms.cell - atoms_ref.cell) < 1e-6
    dist_matrix = get_distances(
        atoms.positions, atoms_ref.positions, cell=atoms_ref.cell, pbc=True)[1]
    permutation = []
    for i in range(len(atoms_ref)):
        dist_row = dist_matrix[:, i]
        permutation.append(np.argmin(dist_row))

    if len(set(permutation)) != len(permutation):
        raise Exception('Duplicates in permutation')
    for i, p in enumerate(permutation):
        if atoms[p].symbol != atoms_ref[i].symbol:
            raise Exception('Matching lattice sites have different occupation')
    return permutation


class Shell:
    """
    Neighbor Shell class

    Parameters
    ----------
    types : list or tuple
        atomic types for neighbor shell
    distance : float
        interatomic distance for neighbor shell
    count : int
        number of pairs in the neighbor shell
    """

    def __init__(self, types, distance, count=0):
        self.types = types
        self.distance = distance
        self.count = count

    def __str__(self):
        s = '{}-{}   distance: {:10.6f}    count: {}'.format(*self.types, self.distance, self.count)
        return s

    __repr__ = __str__


def get_neighbor_shells(atoms: Atoms, cutoff: float, dist_tol: float = 1e-5) -> List[Shell]:
    """ Returns a list of neighbor shells.

    Distances are grouped into shells via the following algorithm:

    1. Find smallest atomic distance `d_min`

    2. Find all pair distances in the range `d_min + 1 * dist_tol`

    3. Construct a shell from these and pop them from distance list

    4. Go to 1.

    Parameters
    ----------
    atoms
        configuration used for finding shells
    cutoff
        exclude neighbor shells which have a distance larger than this value
    dist_tol
        distance tolerance
    """

    # TODO: Remove this once this feature have been in ASE long enough
    try:
        from ase.neighborlist import neighbor_list
    except ImportError:
        raise ImportError('get_neighbor_shells uses new functionality from ASE'
                          ', please update ASE in order to use this function.')

    # get distances
    ijd = neighbor_list('ijd', atoms, cutoff)
    ijd = list(zip(*ijd))
    ijd.sort(key=lambda x: x[2])

    # sort into shells
    symbols = atoms.get_chemical_symbols()
    shells = []
    for i, j, d in ijd:
        types = tuple(sorted([symbols[i], symbols[j]]))
        for shell in shells:
            if abs(d - shell.distance) < dist_tol and types == shell.types:
                shell.count += 1
                break
        else:
            shell = Shell(types, d, 1)
            shells.append(shell)
    shells.sort(key=lambda x: (x.distance, x.types, x.count))

    # warning if two shells are within 2 * tol
    for i, s1 in enumerate(shells):
        for j, s2 in enumerate(shells[i+1:]):
            if s1.types != s2.types:
                continue
            if not s1.distance < s2.distance - 2 * dist_tol:
                logger.warning('Found two shells within 2 * dist_tol')

    return shells


def extract_parameters(fcs: ForceConstants, cs: ClusterSpace):
    """ Extracts parameters from force constants.

    This function can be used to extract parameters to create a
    ForceConstantPotential from a known set of force constants.
    The return values come from NumPy's `lstsq function
    <https://docs.scipy.org/doc/numpy/reference/generated/numpy.linalg.lstsq.html>`_.

    Parameters
    ----------
    fcs
        force constants
    cs
        cluster space
    """
    fcm = ForceConstantModel(fcs.supercell, cs)
    return np.linalg.lstsq(*fcm.get_fcs_sensing(fcs), rcond=None)[0]



""" This module contains functions and variables to control hiPhive's logging

* `logger` - the module logger
* set_config - function to control level and logfile
"""

import logging
import sys
from timeit import default_timer as timer

# This is the root logger of hiPhive
logger = logging.getLogger('hiphive')

# Will process all levels of INFO or higher
logger.setLevel(logging.INFO)

# If you know what you are doing you may set this to True
logger.propagate = False

# The hiPhive logger will collect events from childs and the default behaviour
# is to print it directly to stdout
ch = logging.StreamHandler(sys.stdout)
logger.addHandler(ch)

continuous_logging = False


# TODO: use Context management protocol instead
class Progress:
    """ Progress bar like functionality. """

    def __init__(self, tot=None, mode='frac', estimate_remaining=True):
        if tot is None:
            self._tot = '?'
            assert not estimate_remaining
        else:
            self._tot = tot
        self._progress = 0
        self._estimate_remaining = estimate_remaining
        self._last_update = 0
        self._start = timer()

    def tick(self):
        self._progress += 1
        delta = timer() - self._last_update
        if continuous_logging and delta > 2:
            self._last_update = timer()
            print('\r' + ' ' * 70 + '\r', end='', flush=True)
            print('{}/{}={:.3%}'.format(self._progress,
                  self._tot, self._progress/self._tot), end='', flush=True)
            if self._estimate_remaining and self._tot != '?':
                remaining_time = (self._tot - self._progress) * (
                    timer() - self._start) / self._progress
                print('  time remaining: {:.3}'.format(remaining_time), end='',
                      flush=True)

    def close(self):
        if continuous_logging:
            print('\r' + ' ' * 70 + '\r', end='', flush=True)
        s = timer() - self._start
        d, remainder = divmod(s, 60*60*24)
        h, remainder = divmod(remainder, 60*60)
        m, s = divmod(remainder, 60)
        logger.info('Done in {}d {}h {}m {:.3}s'
                    .format(int(d), int(h), int(m), s))


def set_config(filename=None, level=None, continuous=None):
    """
    Alters the way logging is handled.

    Parameters
    ----------
    filename : str
        name of file the log should be sent to
    level : int
        verbosity level; see `Python logging library
        <https://docs.python.org/3/library/logging.html>`_ for details
    continuous : bool
        if True the progress will be continously updated
    """

    # If a filename is provided a logfile will be created
    if filename is not None:
        fh = logging.FileHandler(filename)
        logger.addHandler(fh)

    if level is not None:
        logger.setLevel(level)

    if continuous is not None:
        global continuous_logging
        continuous_logging = continuous



"""
hiPhive module.
"""

from .cluster_space import ClusterSpace
from .structure_container import StructureContainer
from .force_constant_potential import ForceConstantPotential
from .force_constants import ForceConstants
from .core.config import config
from .core.rotational_constraints import enforce_rotational_sum_rules

# check versions
from warnings import warn
from distutils.version import StrictVersion

from numpy import __version__ as np_version
if StrictVersion(np_version) < StrictVersion('1.12'):
    warn('\n'
         ' The numpy module is outdated (version {}).\n'
         ' hiphive requires at least numpy version 1.12.\n'
         ' Some functionality might not work as expected.\n'
         .format(np_version))

# clean up imports
del np_version
del warn, StrictVersion

__project__ = 'hiPhive'
__description__ = 'High-order force constants for the masses'
__authors__ = ['Fredrik Eriksson',
               'Erik Fransson',
               'Paul Erhart']
__copyright__ = '2019'
__license__ = 'MIT'
__credits__ = ['Fredrik Eriksson',
               'Erik Fransson',
               'Paul Erhart']
__version__ = '0.6'
__all__ = ['ClusterSpace',
           'StructureContainer',
           'ForceConstantPotential',
           'ForceConstants',
           'config',
           'enforce_rotational_sum_rules',
           'io']
__maintainer__ = 'The hiPhive developers team'
__maintainer_email__ = 'hiphive@materialsmodeling.org'
__status__ = 'beta-version'
__url__ = 'http://hiphive.materialsmodeling.org/'



"""
BaseOptimizer serves as base for all optimizers.
"""

import numpy as np
from abc import ABC, abstractmethod
from typing import Any, Dict, Tuple, Union
from .fit_methods import available_fit_methods
from .io import _write_pickle


class BaseOptimizer(ABC):
    """BaseOptimizer class.

    Serves as base class for all Optimizers solving the linear
    :math:`\\boldsymbol{X}\\boldsymbol{a} = \\boldsymbol{y}` problem.

    Parameters
    ----------
    fit_data : tuple(numpy.ndarray, numpy.ndarray)
        the first element of the tuple represents the `NxM`-dimensional
        fit matrix `A` whereas the second element represents the
        vector of `N`-dimensional target values `y`; here `N` (=rows of
        `A`, elements of `y`) equals the number of target values and
        `M` (=columns of `A`) equals the number of parameters
    fit_method : str
        method to be used for training; possible choice are
        "least-squares", "lasso", "elasticnet", "bayesian-ridge", "ardr",
        "rfe", "split-bregman"
    standardize : bool
        if True the fit matrix and target values are standardized before fitting,
        meaning columns in the fit matrix and th target values are rescaled to
        have a standard deviation of 1.0.
    check_condition : bool
        if True the condition number will be checked
        (this can be sligthly more time consuming for larger
        matrices)
    seed : int
        seed for pseudo random number generator
    """

    def __init__(self,
                 fit_data: Tuple[np.ndarray, np.ndarray],
                 fit_method: str,
                 standardize: bool = True,
                 check_condition: bool = True,
                 seed: int = 42):
        """
        Attributes
        ----------
        _A : numpy.ndarray
            fit matrix (N, M)
        _y : numpy.ndarray
            target values (N)
        """

        if fit_method not in available_fit_methods:
            raise ValueError('Unknown fit_method: {}'.format(fit_method))

        if fit_data is None:
            raise TypeError('Invalid fit data; Fit data can not be None')
        if fit_data[0].shape[0] != fit_data[1].shape[0]:
            raise ValueError('Invalid fit data; shapes of fit matrix'
                             ' and target vector do not match')
        if len(fit_data[0].shape) != 2:
            raise ValueError('Invalid fit matrix; must have two dimensions')

        self._A, self._y = fit_data
        self._n_rows = self._A.shape[0]
        self._n_cols = self._A.shape[1]
        self._fit_method = fit_method
        self._standarize = standardize
        self._check_condition = check_condition
        self._seed = seed
        self._fit_results = {'parameters': None}

    def compute_rmse(self, A: np.ndarray, y: np.ndarray) -> float:
        """
        Returns the root mean squared error (RMSE) using
        :math:`\\boldsymbol{A}`, :math:`\\boldsymbol{y}`, and the vector of
        fitted parameters :math:`\\boldsymbol{x}`, corresponding to
        :math:`\\|\\boldsymbol{A}\\boldsymbol{x}-\\boldsymbol{y}\\|_2`.

        Parameters
        ----------
        A
            fit matrix (`N,M` array) where `N` (=rows of `A`, elements
            of `y`) equals the number of target values and `M`
            (=columns of `A`) equals the number of parameters
            (=elements of `x`)
        y
            vector of target values
        """
        y_predicted = self.predict(A)
        delta_y = y_predicted - y
        rmse = np.sqrt(np.mean(delta_y**2))
        return rmse

    def predict(self, A: np.ndarray) -> Union[np.ndarray, float]:
        """
        Predicts data given an input matrix :math:`\\boldsymbol{A}`,
        i.e., :math:`\\boldsymbol{A}\\boldsymbol{x}`, where
        :math:`\\boldsymbol{x}` is the vector of the fitted parameters.
        The method returns the vector of predicted values or a float
        if a single row provided as input.

        Parameters
        ----------
        A
            fit matrix where `N` (=rows of `A`, elements of `y`) equals the
            number of target values and `M` (=columns of `A`) equals the number
            of parameters
        """
        return np.dot(A, self.parameters)

    def get_contributions(self, A: np.ndarray) -> np.ndarray:
        """
        Returns the average contribution for each row of `A`
        to the predicted values from each element of the parameter vector.

        Parameters
        ----------
        A
            fit matrix where `N` (=rows of `A`, elements of `y`) equals the
            number of target values and `M` (=columns of `A`) equals the number
            of parameters
        """
        return np.mean(np.abs(np.multiply(A, self.parameters)), axis=0)

    @abstractmethod
    def train(self) -> None:
        pass

    @property
    def summary(self) -> Dict[str, Any]:
        """ comprehensive information about the optimizer """
        target_values_std = np.std(self._y)

        info = dict()
        info['seed'] = self.seed
        info['fit_method'] = self.fit_method
        info['standardize'] = self.standardize
        info['n_target_values'] = self.n_target_values
        info['n_parameters'] = self.n_parameters
        info['n_nonzero_parameters'] = self.n_nonzero_parameters
        info['parameters_norm'] = self.parameters_norm
        info['target_values_std'] = target_values_std
        return {**info, **self._fit_results}

    def write_summary(self, fname: str):
        """ Writes summary dict to file """
        _write_pickle(fname, self.summary)

    def __str__(self) -> str:
        width = 54
        s = []
        s.append(' {} '.format(self.__class__.__name__).center(width, '='))
        for key in sorted(self.summary.keys()):
            value = self.summary[key]
            if isinstance(value, (str, int, np.integer)):
                s.append('{:30} : {}'.format(key, value))
            elif isinstance(value, (float)):
                s.append('{:30} : {:.7g}'.format(key, value))
        s.append(''.center(width, '='))
        return '\n'.join(s)

    def __repr__(self) -> str:
        return 'BaseOptimizer((A, y), {}, {}'.format(
            self.fit_method, self.seed)

    @property
    def fit_method(self) -> str:
        """ fit method """
        return self._fit_method

    @property
    def parameters(self) -> np.ndarray:
        """ copy of parameter vector """
        if self._fit_results['parameters'] is None:
            return None
        else:
            return self._fit_results['parameters'].copy()

    @property
    def parameters_norm(self) -> float:
        """ the norm of the parameters """
        if self.parameters is None:
            return None
        else:
            return np.linalg.norm(self.parameters)

    @property
    def n_nonzero_parameters(self) -> int:
        """ number of non-zero parameters """
        if self.parameters is None:
            return None
        else:
            return np.count_nonzero(self.parameters)

    @property
    def n_target_values(self) -> int:
        """ number of target values (=rows in `A` matrix) """
        return self._n_rows

    @property
    def n_parameters(self) -> int:
        """ number of parameters (=columns in `A` matrix) """
        return self._n_cols

    @property
    def standardize(self) -> bool:
        """ if True standardize the fit matrix before fitting """
        return self._standarize

    @property
    def seed(self) -> int:
        """ seed used to initialize pseudo random number generator """
        return self._seed



import pickle
import numpy as np
from ase.neighborlist import NeighborList
from hiphive.io.pretty_table_prints import table_array_to_string


# TODO: Use another base class to hide internals from user
class Cutoffs:
    """ This class maintains information about the cutoff configuration,
    i.e. which clusters will be included (="inside cutoff"). It also
    encapsulates functionality that is used e.g., during cluster space
    construction.

    Here, `n-body` refers to number of atoms in a cluster. For example
    the cluster (0011) is a two-body cluster of fourth order and the
    cluster (123) is a three-body cluster of third order.

    Parameters
    ----------
    cutoff_matrix : numpy.ndarray
        the matrix element `ij` provides to the cutoff for order `j+2`
        and nbody `i+2`; elements with `i>j` will be ignored
    """

    def __init__(self, cutoff_matrix):

        try:
            self._cutoff_matrix = np.array(cutoff_matrix, dtype=np.float)
        except Exception:
            raise Exception('Please specify cutoffs as a matrix of numbers')
        if len(self._cutoff_matrix.shape) != 2:
            raise TypeError('Please specify cutoff matrix as a 2D array')
        for i, row in enumerate(self._cutoff_matrix):
            if np.any(row[i:] < 0):
                raise ValueError('Negative number as cutoff')
            row[:i] = np.NaN
        self._cutoff_matrix = self._cutoff_matrix[:(self.max_nbody-1), :(self.max_order-1)]

    @property
    def cutoff_matrix(self):
        """ numpy.ndarray : copy of cutoff matrix """
        return self._cutoff_matrix.copy()

    @property
    def orders(self):
        """ list(int) : allowed orders """
        return list(range(2, self.max_order + 1))

    @property
    def nbodies(self):
        """ list(int) : allowed bodies """
        return list(range(2, self.max_nbody + 1))

    def get_cutoff(self, order, nbody):
        """
        Returns cutoff for a given body and order.

        Parameters
        ----------
        order : int
        nbody : int

        Raises
        ------
        ValueError
            if order is not in orders
        ValueError
            if nbody is not in nbodies
        ValueError
            if nbody is larger than order

        Returns
        -------
        float
        """
        if order not in self.orders:
            raise ValueError('order not in orders')
        if nbody not in self.nbodies:
            raise ValueError('nbody not in nbodies')
        if nbody > order:
            raise ValueError('nbody can not be larger than order')
        return self._cutoff_matrix[nbody - 2, order - 2]

    @property
    def max_cutoff(self):
        """ float : maximum cutoff """
        max_cutoff = 0
        for i, row in enumerate(self._cutoff_matrix):
            max_cutoff = max(max_cutoff, np.max(row[i:]))
        return max_cutoff

    @property
    def max_nbody(self):
        """ int : maximum body """
        nbody = 1
        for i, row in enumerate(self._cutoff_matrix):
            if np.any(row[i:]):
                nbody = i + 2
        return nbody

    @property
    def max_order(self):
        """ int : maximum order """
        order = None
        for col in range(self._cutoff_matrix.shape[1]):
            if np.any(self._cutoff_matrix[:col + 1, col]):
                order = col + 2
        return order

    def max_nbody_cutoff(self, nbody):
        """ Return maximum cutoff for a given body. """
        if nbody not in self.nbodies:
            raise ValueError('nbody not in nbodies')
        return np.max(self._cutoff_matrix[nbody - 2, max(0, nbody - 2):])

    def max_nbody_order(self, nbody):
        """ Returns maximum order for a given body """
        if nbody not in self.nbodies:
            raise ValueError('nbody not in nbodies')
        row = self._cutoff_matrix[nbody - 2]
        max_order = None
        for order, cutoff in enumerate(row[nbody-2:], start=nbody):
            if cutoff:
                max_order = order
        return max_order

    def write(self, fileobj):
        """ Writes instance to file.

        Parameters
        ----------
        fileobj : file-like object
            file-like object to which the cutoffs will be written to
        """
        pickle.dump(self._cutoff_matrix, fileobj)

    def read(fileobj):
        """ Reads an instance from file.

        Parameters
        ----------
        fileobj : file-like object
            input file to read from
        """
        data = pickle.load(fileobj)
        return Cutoffs(data)

    def to_filename_tag(self):
        """ Simple function turning cutoffs into a string to be used in e.g.
        filenames. """
        s = []
        for i, c in enumerate(self._cutoff_matrix.tolist(), start=2):
            s.append('{}body-{}'.format(i, '_'.join(map(str, c))))
        return '_'.join(s)

    def __str__(self):
        cutoff_matrix = self._cutoff_matrix.copy()
        cutoff_matrix = np.vstack(([[None] * len(self.orders)], cutoff_matrix))
        s = table_array_to_string(cutoff_matrix)

        width = max(len(c) for c in s.split('\n'))
        header = ' Cutoffs '.center(width, '=') + '\n'
        bottom = '\n' + ''.center(width, '=')
        s = header + s + bottom
        return s

    def __repr__(self):
        return 'Cutoffs({!r})'.format(self._cutoff_matrix)


class CutoffMaximumBody(Cutoffs):
    """ Specify cutoff-list plus maximum body

    Usefull when creating e.g. 6th order expansions but with only 3-body
    interactions.

    Parameters
    ----------
    cutoff_list : list
        list of cutoffs for order 2, 3, etc. Must be in decresing order
    max_nbody : int
        No clusters containing more than max_nbody atoms will be generated
    """

    def __init__(self, cutoff_list, max_nbody):
        cutoff_matrix = np.zeros((max_nbody - 1, len(cutoff_list)))
        for order, cutoff in enumerate(cutoff_list, start=2):
            cutoff_matrix[:, order - 2] = cutoff
        super().__init__(cutoff_matrix)


def is_cutoff_allowed(atoms, cutoff):
    """ Checks if atoms is compatible with cutoff

    Parameters
    ----------
    atoms : ase.Atoms
        structure used for checking compatibility with cutoff
    cutoff : float
        cutoff to be tested

    Returns
    -------
    bool
        True if cutoff compatible with atoms object, else False
    """
    nbrlist = NeighborList(cutoffs=[cutoff / 2] * len(atoms), skin=0,
                           self_interaction=False, bothways=True)
    nbrlist.update(atoms)

    for i in range(len(atoms)):
        neighbors, _ = nbrlist.get_neighbors(i)
        if i in neighbors:
            return False
        if len(neighbors) != len(set(neighbors)):
            return False
    return True


def estimate_maximum_cutoff(atoms, max_iter=11):
    """ Estimates the maximum possible cutoff given the atoms object

    Parameters
    ----------
    atoms : ase.Atoms
        structure used for checking compatibility with cutoff
    max_iter : int
        number of iterations in binary search
    """

    # First upper boundary of cutoff
    upper_cutoff = min(np.linalg.norm(atoms.cell, axis=1))

    # generate all possible offsets given upper_cutoff
    nbrlist = NeighborList(cutoffs=[upper_cutoff / 2] * len(atoms), skin=0,
                           self_interaction=False, bothways=True)
    nbrlist.update(atoms)
    all_offsets = []
    for i in range(len(atoms)):
        _, offsets = nbrlist.get_neighbors(i)
        all_offsets.extend([tuple(offset) for offset in offsets])

    # find lower boundary and new upper boundary
    unique_offsets = set(all_offsets)
    unique_offsets.discard((0, 0, 0))
    upper = min(np.linalg.norm(np.dot(offset, atoms.cell))
                for offset in unique_offsets)
    lower = upper / 2.0

    # run binary search between the upper and lower bounds
    for _ in range(max_iter):
        cutoff = (upper + lower) / 2
        if is_cutoff_allowed(atoms, cutoff):
            lower = cutoff
        else:
            upper = cutoff
    return lower


class BaseClusterFilter:
    """Base cluster filter class.

    This filter simply accepts all proposed clusters.  A proper
    subclass must implement the same methods.
    """
    def setup(self, atoms):
        """ The filter is passed the environment of the primitive cell.

        Parameters
        ----------
        atoms : ase.Atoms
            non-pbc primitive cell plus neighboring atoms
        """
        self._atoms = atoms

    def __call__(self, cluster):
        """ Returns True or False when a cluster is proposed.

        Parameters
        ----------
        cluster : tuple(int)
            indices of proposed cluster referenced to the internal
            atoms object
        """
        return True


class MaxTripletDistance(BaseClusterFilter):
    """Rejects large three-body clusters.

    If two or all pairs within the cluster are further away from each
    other than the specified distance the cluster is
    rejected. Alternatively, only one pair distance can exceed the
    specified cutoff.

    The filter only rejects for three-body clusters. All other
    clusters are accepted.

    Parameters
    ----------
    max_dist : float
        maximum distance allowed for the two closest pairs

    """
    def __init__(self, max_dist):
        self._max_dist = max_dist

    def __call__(self, cluster):
        cluster = tuple(set(cluster))
        if len(cluster) != 3:
            return True
        i, j, k = cluster
        ij = self._atoms.get_distance(i, j) < self._max_dist
        ik = self._atoms.get_distance(i, k) < self._max_dist
        jk = self._atoms.get_distance(j, k) < self._max_dist

        return (ij and (ik or jk)) or (ik and jk)



"""
This module implements the split-Bregman algorithm described in
T. Goldstein and S. Osher, SIAM J. Imaging Sci. 2, 323 (2009);
doi:10.1137/080725891
"""

import numpy as np
from scipy.optimize import minimize
from typing import Any, Dict


def fit_split_bregman(A: np.ndarray,
                      y: np.ndarray,
                      mu: float = 1e-3,
                      lmbda: float = 100,
                      n_iters: int = 1000,
                      tol: float = 1e-6) -> Dict[str, Any]:
    """
    Determines the solution :math:`\\boldsymbol{x}` to the linear
    problem :math:`\\boldsymbol{A}\\boldsymbol{x}=\\boldsymbol{y}` using
    the split-Bregman algorithm described in T. Goldstein and S. Osher,
    SIAM J. Imaging Sci. 2, 323 (2009); doi:10.1137/080725891.
    The thus obtained parameters are returned in the form of a
    dictionary with a key named `parameters`

    Parameters
    ----------
    A
        fit matrix
    y
        target array
    mu
        sparseness parameter
    lmbda
        weight of additional L2-norm in split-Bregman
    n_iters
        maximal number of split-Bregman iterations
    tol
        convergence criterion iterative minimization
    """

    def _shrink(y: np.ndarray, alpha: float) -> np.ndarray:
        """
        Shrinkage operator as defined in Eq. (11) of the paper by Nelson
        et al., Phys. Rev. B 87, 035125 (2013); doi:10.1103/PhysRevB.87.035125.
        """
        return np.sign(y) * np.maximum(np.abs(y) - alpha, 0.0)

    n_cols = A.shape[1]
    d = np.zeros(n_cols)
    b = np.zeros(n_cols)
    x = np.zeros(n_cols)

    old_norm = 0.0

    # Precompute for speed.
    AtA = np.dot(A.conj().transpose(), A)
    ftA = np.dot(y.conj().transpose(), A)
    ii = 0
    for i in range(n_iters):
        args = (A, y, mu, lmbda, d, b, AtA, ftA)
        res = minimize(_objective_function, x, args, method='BFGS',
                       options={'disp': False},
                       jac=_objective_function_derivative)
        x = res.x

        d = _shrink(mu*x + b, 1.0/lmbda)
        b = b + mu*x - d

        new_norm = np.linalg.norm(x)
        ii = ii + 1

        if abs(new_norm-old_norm) < tol:
            break

        old_norm = new_norm

    fit_results = {'parameters': x}
    return fit_results


def _objective_function(x: np.ndarray, A: np.ndarray, y: np.ndarray,
                        mu: float, lmbda: float, d: np.ndarray, b: np.ndarray,
                        AtA: np.ndarray, ftA: np.ndarray) -> np.ndarray:
    """
    Returns the objective function to be minimized.

    Parameters
    -----------
    X
        fit matrix
    y
        target array
    mu
        the parameter that adjusts sparseness.
    lmbda
        Split Bregman parameter
    d
        same notation as Nelson et al. paper
    b
        same notation as Nelson et al. paper
    AtA
        sensing matrix transpose times sensing matrix.
    ftA
        np.dot(y.conj().transpose(), A)
    """

    error_vector = np.dot(A, x) - y

    obj_function = 0.5*np.vdot(error_vector, error_vector)

    if obj_function.imag > 0.0:
        raise RuntimeError(
            'Objective function contains non-zero imaginary part.)')

    sparseness_correction = d - b - mu*x
    obj_function += 0.5*lmbda * \
        np.vdot(sparseness_correction, sparseness_correction)

    if obj_function.imag > 0.0:
        raise RuntimeError(
            'Objective function contains non-zero imaginary part.)')

    return obj_function


def _objective_function_derivative(x: np.ndarray,
                                   A: np.ndarray,
                                   y: np.ndarray,
                                   mu: float,
                                   lmbda: float,
                                   d: np.ndarray,
                                   b: np.ndarray,
                                   AtA: np.ndarray,
                                   ftA: np.ndarray) -> np.ndarray:
    """
    Returns the derivative of the objective function.

    Parameters
    -----------
    X
        fit matrix
    y
        target array
    mu
        the parameter that adjusts sparseness.
    lmbda
        Split Bregman parameter
    d
        same notation as Nelson, Hart paper
    b
        same notation as Nelson, Hart paper
    AtA
        sensing matrix transpose times sensing matrix.
    ftA
        np.dot(y.conj().transpose(), A)
    """
    ret = np.squeeze(np.dot(x[np.newaxis, :], AtA) -
                     ftA - lmbda*mu*(d - mu * x - b))
    return ret



"""
This module enables the generation of structures with displacements
that can be used to generate reference forces.
"""

from .rattle import (generate_mc_rattled_structures,
                     generate_rattled_structures)
from .phonon import generate_phonon_rattled_structures

__all__ = ['generate_mc_rattled_structures',
           'generate_rattled_structures',
           'generate_phonon_rattled_structures']



"""
Collection of functions and classes for handling information concerning atoms
and structures, including the relationship between primitive cell and
supercells that are derived thereof.
"""

import pickle
from ase import Atoms as aseAtoms
import numpy as np
from ..io.logging import logger

# TODO: Rename logger
logger = logger.getChild('atoms')


class Atom:
    # TODO: This class should inherit some immutable to make it clear that
    # there is no reference to any other obj
    """Unique representation of an atom in a lattice with a basis

    Class for storing information about the position of an atom in a supercell
    relative to the origin of the underlying primitive cell. This class is used
    for handling the relationship between a primitive cell and supercells
    derived thereof.

    Parameters
    ----------
    site : int
        site index
    offset : list(float) or numpy.ndarray
        must contain three elements, offset_x, offset_y, offset_z
    """
    def __init__(self, site, offset):
        offset = tuple(offset)
        self._site = site
        self._offset = offset

    @property
    def site(self):
        """int : index of corresponding site in the primitive basis"""
        return self._site

    @property
    def offset(self):
        """list(int) : translational offset of the supercell site relative
        to the origin of the primitive cell in units of primitive lattice
        vectors"""
        return self._offset

    def __repr__(self):
        return 'Atom({}, {})'.format(self.site, self.offset)

    def spos(atom, basis):
        return np.add(basis[atom.site], atom.offset)

    def pos(atom, basis, cell):
        spos = atom.spos(basis)
        return np.dot(spos, cell)

    @staticmethod
    def spos_to_atom(spos, basis, tol=None):
        # TODO: Why is this simply duplicated spos_to_atom from helper below?
        if not tol:
            # TODO: Link to config file
            tol = 1e-4
        for site, base in enumerate(basis):
            offset = np.subtract(spos, base)
            diff = offset - np.round(offset, 0)
            if np.linalg.norm(diff) < tol:
                offset = np.round(offset, 0).astype(int)
                atom = Atom(site, offset)
                assert np.linalg.norm(spos - atom.spos(basis)) < tol, (
                    '{} with basis {} != {}'.format(atom, basis, spos))
                return atom

        s = '{} not compatible with {} and tolerance {}'
        raise Exception(s.format(spos, basis, tol))

    def __hash__(self):
        return hash((self._site, *self.offset))

    def __eq__(self, other):
        if not isinstance(other, Atom):
            return False
        return self.site == other.site and self.offset == other.offset


class Atoms(aseAtoms):
    """Minimally augmented version of the ASE Atoms class suitable for handling
    primitive cell information.

    Saves and loads by pickle.
    """
    @property
    def basis(self):
        """numpy.ndarray : scaled coordinates of the sites in the primitive basis
        """
        return self.get_scaled_positions().round(12) % 1

    def write(self, f):
        """ Writes the object to file.

        Note: Only the cell, basis and numbers are stored!

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to write to (file object)
        """
        data = {}
        data['cell'] = self.cell
        data['basis'] = self.basis
        data['numbers'] = self.numbers

        try:
            pickle.dump(data, f)
        except Exception:
            raise Exception('Failed writing to file.')

    @staticmethod
    def read(f):
        """ Load an hiPhive Atoms object from file.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to load from (file object)

        Returns
        -------
        hiPhive Atoms object
        """
        try:
            data = pickle.load(f)
        except Exception:
            raise Exception('Failed loading from file.')
        atoms = aseAtoms(numbers=data['numbers'],
                         scaled_positions=data['basis'],
                         cell=data['cell'],
                         pbc=True)
        return Atoms(atoms)


def atom_to_spos(atom, basis):
    """Helper function for obtaining the position of a supercell atom in scaled
    coordinates.

    Parameters
    ----------
    atom : hiPhive.Atom
        supercell atom
    basis : list(list(float)) or numpy.ndarray
        positions of sites in the primitive basis

    Returns
    -------
    numpy.ndarray
        scaled coordinates of an atom in a supercell
    """
    return np.add(atom.offset, basis[atom.site])


def spos_to_atom(spos, basis, tol=1e-4):
    """Helper function for transforming a supercell position to the primitive
    basis.

    Parameters
    ----------
    spos : list(list(float)) or numpy.ndarray
        scaled coordinates of an atom in a supercell
    basis : list(list(float)) or numpy.ndarray
        positions of sites in the primitive basis
    tol : float
        a general tolerance

    Returns
    -------
    hiphive.Atom
         supercell atom
    """
    # TODO: Fix tolerance
    # Loop over all sites in the basis
    for site, base in enumerate(basis):
        # If the scaled position belongs to this site, the offset is the
        # difference in scaled coordinates and should be integer
        offset = np.subtract(spos, base)
        # The diff is the difference between the offset vector and the nearest
        # integer vector.
        diff = offset - np.round(offset, 0)
        # It should be close to the null vector if this is the correct site.
        if np.linalg.norm(diff) < tol:
            # If the difference is less than the tol make the offset integers
            offset = np.round(offset, 0).astype(int)
            # This should be the correct atom
            atom = Atom(site, offset)
            # Just to be sure we check that the atom actually produces the
            # input spos given the input basis
            s = ('Atom=[{},{}] with basis {} != {}'
                 .format(atom.site, atom.offset, basis, spos))
            assert np.linalg.norm(spos - atom_to_spos(atom, basis)) < tol, s
            return atom
    # If no atom was found we throw an error
    s = '{} not compatible with {} and tolerance {}'.format(spos, basis, tol)
    # TODO: Should we throw more explicit error?
    raise Exception(s)
